import logging

from django.http import JsonResponse
from django.views import View

from djangoldp.endpoints.webfinger import WebFingerEndpoint, WebFingerError

logger = logging.getLogger('djangoldp')


class WebFingerView(View):
    endpoint_class = WebFingerEndpoint

    def get(self, request, *args, **kwargs):
        return self.on_request(request)

    def on_request(self, request):
        endpoint = self.endpoint_class(request)
        try:
            endpoint.validate_params()

            return JsonResponse(endpoint.response())

        except WebFingerError as error:
            return JsonResponse(error.create_dict(), status=400)

    def post(self, request, *args, **kwargs):
        return self.on_request(request)


import logging
from collections import OrderedDict

from django.conf import settings
from pyld import jsonld
from rest_framework.authentication import SessionAuthentication
from rest_framework.exceptions import ParseError
from rest_framework.parsers import JSONParser
from rest_framework.renderers import JSONRenderer

logger = logging.getLogger('djangoldp')


# renders into JSONLD format by applying context to the data
# https://github.com/digitalbazaar/pyld
class JSONLDRenderer(JSONRenderer):
    media_type = 'application/ld+json'

    def render(self, data, accepted_media_type=None, renderer_context=None):
        if isinstance(data, dict):
            context = data.get("@context")
            if isinstance(context, list):
                context_value = [settings.LDP_RDF_CONTEXT] + context
            elif isinstance(context, str) or isinstance(context, dict):
                context_value = [settings.LDP_RDF_CONTEXT, context]
            else:
                context_value = settings.LDP_RDF_CONTEXT

            ordered_data = OrderedDict()
            ordered_data["@context"] = context_value
            for key, value in data.items():
                if key != "@context":
                    ordered_data[key] = value
            data = ordered_data

        return super(JSONLDRenderer, self).render(data, accepted_media_type, renderer_context)


# https://github.com/digitalbazaar/pyld
class JSONLDParser(JSONParser):
    #TODO: It current only works with pyld 1.0. We need to check our support of JSON-LD
    media_type = 'application/ld+json'

    def parse(self, stream, media_type=None, parser_context=None):
        data = super(JSONLDParser, self).parse(stream, media_type, parser_context)
        # compact applies the context to the data and makes it a format which is easier to work with
        # see: http://json-ld.org/spec/latest/json-ld/#compacted-document-form
        try:
            return jsonld.compact(data, ctx=settings.LDP_RDF_CONTEXT)
        except jsonld.JsonLdError as e:
            raise ParseError(str(e.cause))


# an authentication class which exempts CSRF authentication
class NoCSRFAuthentication(SessionAuthentication):
    def enforce_csrf(self, request):
        return




from csv import DictWriter

from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from django.core.exceptions import FieldDoesNotExist
from django.http import HttpResponse
from guardian.admin import GuardedModelAdmin

from djangoldp.activities.services import ActivityQueueService
from djangoldp.models import Activity, Follower, ScheduledActivity


class DjangoLDPAdmin(GuardedModelAdmin):
    '''
    An admin model representing a federated object. Inherits from GuardedModelAdmin to provide Django-Guardian
    object-level permissions
    '''
    actions = ['export_csv']
    export_fields = []

    def resolve_verbose_name(self, field_path):
        field = self
        for field_name in field_path.split('__'):
            try:
                field = field.model._meta.get_field(field_name)
            except FieldDoesNotExist:
                return None
        return field.verbose_name

    @admin.action(description="Export CSV")
    def export_csv(self, request, queryset):
        response = HttpResponse(content_type="text/csv")
        response['Content-Disposition'] = f'attachment; filename="{self.model.__name__}.csv"'
        # only keep fields that can be resolved, keep only urlid if none
        field_list = list(filter(self.resolve_verbose_name, self.export_fields or self.list_display)) or ['urlid']
        headers = {field:self.resolve_verbose_name(field) for field in field_list}

        writer = DictWriter(response, fieldnames=field_list)
        writer.writerow(headers)
        writer.writerows(queryset.values(*field_list))
        return response


class DjangoLDPUserAdmin(UserAdmin, DjangoLDPAdmin):
    '''An extension of UserAdmin providing the functionality of DjangoLDPAdmin'''

    list_display = ('urlid', 'email', 'first_name', 'last_name', 'date_joined', 'last_login', 'is_staff')
    search_fields = ['urlid', 'email', 'first_name', 'last_name']
    ordering = ['urlid']

    def get_fieldsets(self, request, obj=None):
        fieldsets = super().get_fieldsets(request, obj)
        
        federated_fields = ['urlid', 'allow_create_backlink']
        if self.exclude is not None:
            federated_fields = list(set(federated_fields) - set(self.exclude))

        for fieldset in fieldsets:
            federated_fields = list(set(federated_fields) - set(fieldset[1]['fields']))

        if len(federated_fields) == 0:
            return fieldsets

        fieldsets = [('Federation', {'fields': federated_fields})] + list(fieldsets)

        return fieldsets


@admin.action(description='Resend activity')
def resend_activity(modeladmin, request, queryset):
    for a in queryset:
        ActivityQueueService.send_activity(a.external_id, a.to_activitystream())
resend_activity.short_description = 'Resend activity'


@admin.register(Activity, ScheduledActivity)
class ActivityAdmin(DjangoLDPAdmin):
    fields = ['urlid', 'type', 'local_id', 'external_id', 'created_at', 'success', 'payload_view', 'response_code',
              'response_location', 'response_body']
    list_display = ['created_at', 'type', 'local_id', 'external_id', 'success', 'response_code']
    readonly_fields = ['created_at', 'payload_view', 'response_location', 'response_code', 'response_body']
    search_fields = ['urlid', 'type', 'local_id', 'external_id', 'response_code']
    actions = [resend_activity]

    def payload_view(self, obj):
        return str(obj.to_activitystream())


@admin.register(Follower)
class FollowerAdmin(DjangoLDPAdmin):
    fields = ['urlid', 'object', 'inbox', 'follower']
    list_display = ['urlid', 'object', 'inbox', 'follower']
    search_fields = ['object', 'inbox', 'follower']





import uuid
import json
from collections import OrderedDict
from collections.abc import Mapping, Iterable
from copy import copy
from typing import Any
from urllib import parse

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.core.exceptions import ValidationError as DjangoValidationError
from django.db import transaction
from django.db.models import QuerySet
from django.urls import resolve, Resolver404, get_script_prefix
from django.urls.resolvers import get_resolver
from django.utils.datastructures import MultiValueDictKeyError
from django.utils.encoding import uri_to_iri
from django.utils.functional import cached_property
from rest_framework.exceptions import ValidationError
from rest_framework.fields import SkipField, empty, ReadOnlyField
from rest_framework.fields import get_error_detail, set_value
from rest_framework.relations import HyperlinkedRelatedField, ManyRelatedField, Hyperlink, MANY_RELATION_KWARGS
from rest_framework.serializers import HyperlinkedModelSerializer, ListSerializer, ModelSerializer, LIST_SERIALIZER_KWARGS
from rest_framework.settings import api_settings
from rest_framework.utils import model_meta
from rest_framework.utils.field_mapping import get_nested_relation_kwargs
from rest_framework.utils.serializer_helpers import ReturnDict, BindingDict

from djangoldp.fields import LDPUrlField, IdURLField
from djangoldp.models import Model
from djangoldp.permissions import DEFAULT_DJANGOLDP_PERMISSIONS

# defaults for various DjangoLDP settings (see documentation)
MAX_RECORDS_SERIALIZER_CACHE = getattr(settings, 'MAX_RECORDS_SERIALIZER_CACHE', 10000)

class InMemoryCache:
    def __init__(self):
        self.cache = {
        }

    def reset(self):
        self.cache = {
        }

    def has(self, cache_key, container_urlid=None, vary=None):
        return cache_key in self.cache and \
               (container_urlid is None or container_urlid in self.cache[cache_key]) and \
               (vary is None or vary in self.cache[cache_key][container_urlid])

    def get(self, cache_key, container_urlid, vary):
        if self.has(cache_key, container_urlid, vary):
            return self.cache[cache_key][container_urlid][vary]['value']
        else:
            return None

    def set(self, cache_key, container_urlid, vary, value):
        if len(self.cache.keys()) > MAX_RECORDS_SERIALIZER_CACHE:
            self.reset()
        
        if cache_key not in self.cache:
            self.cache[cache_key] = {}
        if container_urlid not in self.cache[cache_key]:
            self.cache[cache_key][container_urlid] = {}
        self.cache[cache_key][container_urlid][vary] = {'value': value}

    def invalidate(self, cache_key, container_urlid=None, vary=None):
        # can clear cache_key -> container_urlid -> vary, cache_key -> container_urlid or cache_key
        if container_urlid is not None:
            if vary is not None:
                self.cache[cache_key][container_urlid].pop(vary, None)
            else:
                self.cache[cache_key].pop(container_urlid, None)
        else:
            self.cache.pop(cache_key, None)
GLOBAL_SERIALIZER_CACHE = InMemoryCache()

class RDFSerializerMixin:
    def add_permissions(self, data, user, model, obj=None):
        '''takes a set or list of permissions and returns them in the JSON-LD format'''
        if self.parent and not settings.LDP_INCLUDE_INNER_PERMS: #Don't serialize permissions on nested objects
            return data

        if user.is_superuser:
            data['permissions'] = getattr(settings, 'DJANGOLDP_PERMISSIONS', DEFAULT_DJANGOLDP_PERMISSIONS)
            return data

        permission_classes = getattr(model._meta, 'permission_classes', [])
        if not permission_classes:
            return data
        # The permissions must be given by all permission classes to be granted
        permissions = set.intersection(*[permission().get_permissions(user, model, obj) for permission in permission_classes])
        # Don't grant delete permissions on containers
        if not obj and 'delete' in permissions:
            permissions.remove('delete')
        data['permissions'] = permissions
        return data
    
    def serialize_rdf_fields(self, obj, data, include_context=False):
        '''adds the @type and the @context to the data'''
        rdf_type = getattr(obj._meta, 'rdf_type', None)
        rdf_context = getattr(obj._meta, 'rdf_context', None)
        if rdf_type:
            data['@type'] = rdf_type
        if include_context and rdf_context:
            data['@context'] = rdf_context
        return data

    def serialize_container(self, data, id, user, model, obj=None):
        '''turns a list into a container representation'''
        return self.add_permissions({'@id': id, '@type': 'ldp:Container', 'ldp:contains': data}, user, model, obj)

class LDListMixin(RDFSerializerMixin):
    '''A Mixin for serializing containers into JSONLD format'''
    child_attr = 'child'
    with_cache = getattr(settings, 'SERIALIZER_CACHE', True)

    def get_child(self):
        return getattr(self, self.child_attr)
    
    # converts primitive data representation to the representation used within our application
    def to_internal_value(self, data):
        try:
            # if this is a container, the data will be stored in ldp:contains
            data = data['ldp:contains']
        except (TypeError, KeyError):
            pass

        if len(data) == 0:
            return []
        if isinstance(data, dict):
            data = [data]
        if isinstance(data, str) and data.startswith("http"):
            data = [{'@id': data}]

        return [self.get_child().to_internal_value(item) for item in data]
    
    def filter_queryset(self, queryset, child_model):
        '''Applies the permission of the child model to the child queryset'''
        view = copy(self.context['view'])
        view.model = child_model
        filter_backends = list({perm_class().get_filter_backend(child_model) for perm_class in 
                getattr(child_model._meta, 'permission_classes', []) if hasattr(perm_class(), 'get_filter_backend')})
        for backend in filter_backends:
            if backend:
                queryset = backend().filter_queryset(self.context['request'], queryset, view)
        return queryset
    
    def compute_id(self, value):
        '''generates the @id of the container'''
        if not hasattr(self, 'parent_instance'):
            #This is a container
            return f"{settings.BASE_URL}{self.context['request'].path}"
        return f"{settings.BASE_URL}{Model.resource_id(self.parent_instance)}{self.field_name}/"
    
    def get_attribute(self, instance):
        # save the parent object for nested field url
        self.parent_instance = instance
        return super().get_attribute(instance)

    def check_cache(self, value, id, model, cache_vary):
        '''Auxiliary function to avoid code duplication - checks cache and returns from it if it has entry'''
        parent_meta = getattr(self.get_child(), 'Meta', getattr(self.parent, 'Meta', None))
        depth = max(getattr(parent_meta, "depth", 0), 0) if parent_meta else 1
        
        if depth:
            # if the depth is greater than 0, we don't hit the cache, because a nested container might be outdated
            # this Mixin may not have access to the depth of the parent serializer, e.g. if it's a ManyRelatedField
            # in these cases we assume the depth is 0 and so we hit the cache
            return None
        cache_key = getattr(model._meta, 'label', None)

        if self.with_cache and GLOBAL_SERIALIZER_CACHE.has(cache_key, id, cache_vary):
            cache_value = GLOBAL_SERIALIZER_CACHE.get(cache_key, id, cache_vary)
            # this check is to handle the situation where the cache has been invalidated by something we don't check
            # namely if my permissions are upgraded then I may have access to view more objects
            cache_under_value = cache_value['ldp:contains'] if 'ldp:contains' in cache_value else cache_value
            if not hasattr(cache_under_value, '__len__') or not hasattr(value, '__len__') or (len(cache_under_value) == len(value)):
                return cache_value
        return False

    def to_representation(self, value):
        '''
        Converts internal representation to primitive data representation
        Filters objects out which I don't have permission to view
        Permission on container :
         - Can Add if add permission on contained object's type
         - Can view the container is view permission on container model : container obj are filtered by view permission
        '''
        try:
            child_model = self.get_child().Meta.model
        except AttributeError:
            child_model = value.model
        user = self.context['request'].user
        id = self.compute_id(value)
        
        is_container = True
        if getattr(self, 'parent', None): #If we're in a nested container
            if isinstance(value, QuerySet) and getattr(self, 'parent', None):
                value = self.filter_queryset(value, child_model)

            if getattr(self, 'field_name', None) is not None:
                if self.field_name in getattr(self.parent.Meta.model._meta, 'empty_containers', []):
                    return {'@id': id}
                if not self.field_name in getattr(self.parent.Meta.model._meta, 'nested_fields', []):
                    is_container = False

        cache_vary = str(user)
        cache_result = self.check_cache(value, id, child_model, cache_vary)
        if cache_result:
            return cache_result
        
        data = super().to_representation(value)
        if is_container:
            data = self.serialize_container(data, id, user, child_model)

        GLOBAL_SERIALIZER_CACHE.set(getattr(child_model._meta, 'label'), id, cache_vary, data)
        return GLOBAL_SERIALIZER_CACHE.get(getattr(child_model._meta, 'label'), id, cache_vary)

    def get_value(self, dictionary):
        try:
            object_list = dictionary['@graph']

            if self.parent.instance is None:
                obj = next(filter(
                    lambda o: not hasattr(o, self.parent.url_field_name) or "./" in o[self.parent.url_field_name],
                    object_list))
            else:
                container_id = Model.container_id(self.parent.instance)
                obj = next(filter(lambda o: container_id.lstrip('/') in o[self.parent.url_field_name], object_list))
            list = super().get_value(obj)
            try:
                list = next(
                    filter(lambda o: list[self.parent.url_field_name] == o[self.parent.url_field_name], object_list))
            except (KeyError, TypeError, StopIteration):
                pass

            try:
                list = list['ldp:contains']
            except (KeyError, TypeError):
                pass

            if list is empty:
                return []

            if isinstance(list, dict):
                list = [list]

            ret = []
            for item in list:
                full_item = None
                try:
                    full_item = next(filter(
                        lambda o: self.parent.url_field_name in o and item[self.parent.url_field_name] == o[
                            self.parent.url_field_name], object_list))
                except StopIteration:
                    pass
                if full_item is None:
                    ret.append(item)
                else:
                    ret.append(full_item)

            return ret
        except KeyError:
            obj = super().get_value(dictionary)
            if isinstance(obj, dict) and self.parent.url_field_name in obj:
                resource_id = obj[self.parent.url_field_name]
                if isinstance(resource_id, str) and resource_id.startswith("_:"):
                    object_list = self.root.initial_data['@graph']
                    obj = [next(filter(lambda o: resource_id in o[self.parent.url_field_name], object_list))]

            return obj

class IdentityFieldMixin:
    def to_internal_value(self, data):
        '''Gives the @id as a representation if present'''
        try:
            return super().to_internal_value(data[self.parent.url_field_name])
        except (KeyError, TypeError):
            return super().to_internal_value(data)

class ContainerSerializer(LDListMixin, ListSerializer, IdentityFieldMixin):
    id = ''

    @property
    def data(self):
        return ReturnDict(super(ListSerializer, self).data, serializer=self)


class ManyJsonLdRelatedField(LDListMixin, ManyRelatedField):
    child_attr = 'child_relation'
    url_field_name = "@id"


class JsonLdField(HyperlinkedRelatedField, IdentityFieldMixin):
    def __init__(self, view_name=None, **kwargs):
        super().__init__(view_name, **kwargs)
        self.get_lookup_args()

    def get_url(self, obj, view_name, request, format):
        '''Overridden from DRF to shortcut on urlid-holding objects'''
        if hasattr(obj, 'urlid') and obj.urlid not in (None, ''):
            return obj.urlid
        return super().get_url(obj, view_name, request, format)

    def get_lookup_args(self):
        try:
            lookup_field = get_resolver().reverse_dict[self.view_name][0][0][1][0]
            self.lookup_field = lookup_field
            self.lookup_url_kwarg = lookup_field
        except MultiValueDictKeyError:
            pass

class JsonLdRelatedField(JsonLdField, RDFSerializerMixin):
    def use_pk_only_optimization(self):
        return False

    def to_representation(self, value):
        try:
            include_context = False
            if Model.is_external(value):
                data = {'@id': value.urlid}
            else:
                include_context = True
                data = {'@id': super().to_representation(value)}
            return self.serialize_rdf_fields(value, data, include_context=include_context)
        except ImproperlyConfigured:
            return value.pk

    @classmethod
    def many_init(cls, *args, **kwargs):
        list_kwargs = {'child_relation': cls(*args, **kwargs),}
        for key in kwargs:
            if key in MANY_RELATION_KWARGS:
                list_kwargs[key] = kwargs[key]
        return ManyJsonLdRelatedField(**list_kwargs)


class JsonLdIdentityField(JsonLdField):
    '''Represents an identity (url) field for a serializer'''
    def __init__(self, view_name=None, **kwargs):
        kwargs['read_only'] = True
        kwargs['source'] = '*'
        super().__init__(view_name, **kwargs)

    def use_pk_only_optimization(self):
        return False

    def to_representation(self, value: Any) -> Any:
        '''returns hyperlink representation of identity field'''
        try:
            # we already have a url to return
            if isinstance(value, str):
                return Hyperlink(value, value)
            # expecting a user instance. Compute the webid and return this in hyperlink format
            else:
                return Hyperlink(value.urlid, value)
        except AttributeError:
            return super().to_representation(value)

    def get_attribute(self, instance):
        if Model.is_external(instance):
            return instance.urlid
        else:
            # runs DRF's RelatedField.get_attribute
            # returns the pk only if optimised or the instance itself in the standard case
            return super().get_attribute(instance)


class LDPSerializer(HyperlinkedModelSerializer, RDFSerializerMixin):
    url_field_name = "@id"
    serializer_related_field = JsonLdRelatedField
    serializer_url_field = JsonLdIdentityField
    ModelSerializer.serializer_field_mapping[LDPUrlField] = IdURLField

    # The default serializer repr ends in infinite loop. Overloading it prevents that.
    def __repr__(self):
        return self.__class__.__name__

    @cached_property
    def fields(self):
        """
        A dictionary of {field_name: field_instance}.
        """
        # `fields` is evaluated lazily. We do this to ensure that we don't
        # have issues importing modules that use ModelSerializers as fields,
        # even if Django's app-loading stage has not yet run.
        fields = BindingDict(self)

        # we allow the request object to specify a subset of fields which should be serialized
        model_fields = self.get_fields()
        req_header_accept_shape = self.context['request'].META.get('HTTP_ACCEPT_MODEL_FIELDS') if 'request' in self.context else None
        try:
            allowed_fields = list(set(json.loads(req_header_accept_shape)).intersection(model_fields.keys())) if req_header_accept_shape is not None else model_fields.keys()
        except json.decoder.JSONDecodeError:
            raise ValidationError("Please send the HTTP header Accept-Model-Fields as an array of strings")

        for key, value in model_fields.items():
            if key in allowed_fields:
                fields[key] = value
        
        return fields

    def get_default_field_names(self, declared_fields, model_info):
        try:
            fields = list(self.Meta.model._meta.serializer_fields)
        except AttributeError:
            fields = super().get_default_field_names(declared_fields, model_info)
        return fields + list(getattr(self.Meta, 'extra_fields', []))

    def to_representation(self, obj):
        # external Models should only be returned with rdf values
        if Model.is_external(obj):
            data = {'@id': obj.urlid}
            return self.serialize_rdf_fields(obj, data)

        data = super().to_representation(obj)

        slug_field = Model.slug_field(obj)
        for field in data:
            if isinstance(data[field], dict) and '@id' in data[field]:
                data[field]['@id'] = data[field]['@id'].format(Model.container_id(obj), str(getattr(obj, slug_field)))
        # prioritise urlid field over generated @id
        if 'urlid' in data and data['urlid'] is not None:
            data['@id'] = data.pop('urlid')['@id']
        if not '@id' in data:
            data['@id'] = '{}{}'.format(settings.SITE_URL, Model.resource(obj))

        data = self.serialize_rdf_fields(obj, data, include_context=True)
        data = self.add_permissions(data, self.context['request'].user, type(obj), obj=obj)
        return data

    def build_property_field(self, field_name, model_class):
        class JSonLDPropertyField(ReadOnlyField):
            def to_representation(self, instance):
                from djangoldp.views.ldp_viewset import LDPViewSet
                if isinstance(instance, QuerySet):
                    model = instance.model
                elif isinstance(instance, Model):
                    model = type(instance)
                else:
                    return instance
                depth = max(getattr(self.parent.Meta, "depth", 0) - 1, 0)
                fields = ["@id"] if depth==0 else getattr(model._meta, 'serializer_fields', [])

                serializer_generator = LDPViewSet(model=model, fields=fields, depth=depth,
                                                    lookup_field=getattr(model._meta, 'lookup_field', 'pk'),
                                                    permission_classes=getattr(model._meta, 'permission_classes', []),
                                                    nested_fields=getattr(model._meta, 'nested_fields', []))
                serializer = serializer_generator.get_serializer_class()(context=self.parent.context)

                if isinstance(instance, QuerySet):
                    id = '{}{}{}/'.format(settings.SITE_URL, '{}{}/', self.source)
                    children = [serializer.to_representation(item) for item in instance]
                    return self.parent.serialize_container(children, id, self.parent.context['request'].user, model)
                else:
                    return serializer.to_representation(instance)

        return JSonLDPropertyField, {}

    def handle_value_object(self, value):
        '''
        In JSON-LD value-objects can be passed in, which store some context on the field passed. By overriding this
        function you can react to this context on a field without overriding build_standard_field
        '''
        return value['@value']

    def build_standard_field(self, field_name, model_field):
        class JSonLDStandardField:
            parent_view_name = None

            def __init__(self, **kwargs):
                self.parent_view_name = kwargs.pop('parent_view_name')
                super().__init__(**kwargs)

            def get_value(self, dictionary):
                if self.field_name == 'urlid':
                    self.field_name = '@id'
                try:
                    object_list = dictionary["@graph"]
                    if self.parent.instance is None:
                        obj = next(filter(
                            lambda o: not self.parent.url_field_name in o or "./" in o[self.parent.url_field_name],
                            object_list))
                        value = super().get_value(obj)
                    else:
                        resource_id = Model.resource_id(self.parent.instance)
                        obj = next(
                            filter(lambda o: resource_id.lstrip('/') in o[self.parent.url_field_name], object_list))
                        value = super().get_value(obj)
                except KeyError:
                    value = super().get_value(dictionary)

                if self.field_name == '@id' and value == './':
                    self.field_name = 'urlid'
                    return None

                if isinstance(value, dict) and '@value' in value:
                    value = self.parent.handle_value_object(value)

                return self.manage_empty(value)

            def manage_empty(self, value):
                if value == '' and self.allow_null:
                    # If the field is blank, and null is a valid value then
                    # determine if we should use null instead.
                    return '' if getattr(self, 'allow_blank', False) else None
                elif value == '' and not self.required:
                    # If the field is blank, and emptiness is valid then
                    # determine if we should use emptiness instead.
                    return '' if getattr(self, 'allow_blank', False) else empty
                return value

        field_class, field_kwargs = super().build_standard_field(field_name, model_field)
        field_kwargs['parent_view_name'] = '{}-list'.format(model_field.model._meta.object_name.lower())
        return type(field_class.__name__ + 'Valued', (JSonLDStandardField, field_class), {}), field_kwargs

    def build_nested_field(self, field_name, relation_info, nested_depth):
        class NestedLDPSerializer(self.__class__):
            class Meta:
                model = relation_info.related_model
                depth = nested_depth - 1
                try:
                    fields = ['@id'] + list(model._meta.serializer_fields)
                except AttributeError:
                    fields = '__all__'

            def to_internal_value(self, data):
                if data == '':
                    return ''
                if self.url_field_name in data:
                    if not isinstance(data, Mapping):
                        message = self.error_messages['invalid'].format(
                            datatype=type(data).__name__
                        )
                        raise ValidationError({
                            api_settings.NON_FIELD_ERRORS_KEY: [message]
                        }, code='invalid')

                    ret = OrderedDict()
                    errors = OrderedDict()

                    # validate fields passed in the data
                    fields = list(filter(lambda x: x.field_name in data, self._writable_fields))
                    for field in fields:
                        validate_method = getattr(self, 'validate_' + field.field_name, None)
                        primitive_value = field.get_value(data)
                        try:
                            validated_value = field.run_validation(primitive_value)
                            if validate_method is not None:
                                validated_value = validate_method(validated_value)
                        except ValidationError as exc:
                            errors[field.field_name] = exc.detail
                        except DjangoValidationError as exc:
                            errors[field.field_name] = get_error_detail(exc)
                        except SkipField:
                            pass
                        else:
                            set_value(ret, field.source_attrs, validated_value)

                    if errors:
                        raise ValidationError(errors)

                    # if it's a local resource - use the path to resolve the slug_field on the model
                    uri = data[self.url_field_name]
                    if not Model.is_external(uri):
                        http_prefix = uri.startswith(('http:', 'https:'))

                        if http_prefix:
                            uri = parse.urlparse(uri).path
                            prefix = get_script_prefix()
                            if uri.startswith(prefix):
                                uri = '/' + uri[len(prefix):]

                        try:
                            match = resolve(uri_to_iri(uri))
                            slug_field = Model.slug_field(self.__class__.Meta.model)
                            ret[slug_field] = match.kwargs[slug_field]
                        except Resolver404:
                            pass

                    if 'urlid' in data:
                        ret['urlid'] = data['urlid']

                else:
                    ret = super().to_internal_value(data)

                # copy url_field_name value to urlid, if necessary
                if self.url_field_name in data and not 'urlid' in data and data[self.url_field_name].startswith('http'):
                    ret['urlid'] = data[self.url_field_name]

                return ret

        kwargs = get_nested_relation_kwargs(relation_info)
        kwargs['read_only'] = False
        kwargs['required'] = False
        return NestedLDPSerializer, kwargs

    @classmethod
    def many_init(cls, *args, **kwargs):
        allow_empty = kwargs.pop('allow_empty', None)
        child_serializer = cls(*args, **kwargs)
        list_kwargs = {
            'child': child_serializer,
        }
        if allow_empty is not None:
            list_kwargs['allow_empty'] = allow_empty
        list_kwargs.update({
            key: value for key, value in kwargs.items()
            if key in LIST_SERIALIZER_KWARGS
        })
        meta = getattr(cls, 'Meta', None)
        list_serializer_class = getattr(meta, 'list_serializer_class', ContainerSerializer)
        serializer = list_serializer_class(*args, **list_kwargs)

        # if the child serializer has disabled the cache, really it means disable it on the container
        if hasattr(child_serializer, 'with_cache'):
            serializer.with_cache = child_serializer.with_cache

        return serializer

    def to_internal_value(self, data):
        #TODO: This hack is needed because external users don't pass validation.
        # Objects require all fields to be optional to be created as external, and username is required.
        is_user_and_external = self.Meta.model is get_user_model() and '@id' in data and Model.is_external(data['@id'])
        if is_user_and_external:
            data['username'] = 'external'
        ret = super().to_internal_value(data)
        if is_user_and_external:
            ret['urlid'] = data['@id']
            ret.pop('username')
        return ret

    def get_value(self, dictionary):
        '''overrides get_value to handle @graph key'''
        try:
            object_list = dictionary["@graph"]
            if self.parent.instance is None:
                obj = next(filter(
                    lambda o: not hasattr(o, self.parent.url_field_name) or "./" in o[self.url_field_name],
                    object_list))
            else:
                container_id = Model.container_id(self.parent.instance)
                obj = next(filter(lambda o: container_id.lstrip('/') in o[self.url_field_name], object_list))
            item = super().get_value(obj)
            full_item = None
            if item is empty:
                return empty
            try:
                full_item = next(
                    filter(lambda o: self.url_field_name in o and (item[self.url_field_name] == o[self.url_field_name]),
                           object_list))
            except StopIteration:
                pass
            if full_item is None:
                return item
            else:
                return full_item

        except KeyError:
            return super().get_value(dictionary)

    def create(self, validated_data):
        with transaction.atomic():
            instance = self.internal_create(validated_data, model=self.Meta.model)
            self.attach_related_object(instance, validated_data)

        return instance

    def attach_related_object(self, instance, validated_data):
        '''adds m2m relations included in validated_data to the instance'''
        model_class = self.Meta.model

        info = model_meta.get_field_info(model_class)
        for field_name, relation_info in info.relations.items():
            if relation_info.to_many and relation_info.reverse and not field_name is None:
                rel = getattr(instance._meta.model, field_name).rel
                if rel.name in validated_data:
                    related = validated_data[rel.name]
                    getattr(instance, field_name).add(related)

    def internal_create(self, validated_data, model):
        validated_data = self.resolve_fk_instances(model, validated_data, True)

        # build tuples list of nested_field keys and their values. All list values are considered nested fields
        nested_fields = []
        nested_list_fields_name = list(filter(lambda key: isinstance(validated_data[key], list), validated_data))
        for field_name in nested_list_fields_name:
            nested_fields.append((field_name, validated_data.pop(field_name)))

        info = model_meta.get_field_info(model)
        many_to_many = []
        one_to_one = {}
        for field_name, relation_info in info.relations.items():
            if relation_info.to_many and relation_info.reverse and (
                    field_name in validated_data) and not field_name is None:
                many_to_many.append((field_name, validated_data.pop(field_name)))
            elif relation_info.reverse and (field_name in validated_data) and not field_name is None:
                one_to_one[field_name] = validated_data[field_name]
        validated_data = self.remove_empty_value(validated_data)

        if model is get_user_model() and not 'username' in validated_data:
            validated_data['username'] = str(uuid.uuid4())
        instance = model.objects.create(**validated_data)

        for field_name, value in many_to_many:
            validated_data[field_name] = value

        if one_to_one:
             for field_name, value in one_to_one.items():
                 setattr(instance, field_name, value)
                 value.save()

        self.save_or_update_nested_list(instance, nested_fields)

        return instance

    def remove_empty_value(self, validated_data):
        '''sets any empty strings in the validated_data to None'''
        for attr, value in validated_data.items():
            if value == '':
                validated_data[attr] = None
        return validated_data

    def update(self, instance, validated_data):
        model = self.Meta.model
        nested_fields = []
        nested_fields_name = list(filter(lambda key: isinstance(validated_data[key], list), validated_data))
        for field_name in nested_fields_name:
            nested_fields.append((field_name, validated_data.pop(field_name)))

        for attr, value in validated_data.items():
            if isinstance(value, dict):
                value = self.update_dict_value(attr, instance, value)
            if value == '' and not isinstance(getattr(instance, attr), str):
                setattr(instance, attr, None)
            else:
                setattr(instance, attr, value)

        self.save_or_update_nested_list(instance, nested_fields)
        instance.save()

        return instance

    def resolve_fk_instances(self, model, validated_data, create=False):
        '''
        iterates over every dict object in validated_data and resolves them into instances (get or create)
        :param model: the model being operated on
        :param validated_data: the data passed to the serializer
        :param create: set to True, foreign keys will be created if they do not exist
        '''
        nested_fk_fields_name = list(filter(lambda key: isinstance(validated_data[key], dict), validated_data))
        for field_name in nested_fk_fields_name:
            field_dict = validated_data[field_name]
            field_model = model._meta.get_field(field_name).related_model

            slug_field = Model.slug_field(field_model)
            sub_inst = None
            if 'urlid' in field_dict:
                # has urlid and is a local resource
                model, sub_inst = self.get_inst_by_urlid(field_dict, field_model, model, slug_field, sub_inst)
            # try slug field, assuming that this is a local resource
            elif slug_field in field_dict:
                kwargs = {slug_field: field_dict[slug_field]}
                sub_inst = field_model.objects.get(**kwargs)
            if sub_inst is None:
                if create:
                    sub_inst = self.internal_create(field_dict, field_model)
                else:
                    continue

            validated_data[field_name] = sub_inst
        return validated_data

    def get_inst_by_urlid(self, field_dict, field_model, model, slug_field, sub_inst):
        if parse.urlparse(settings.BASE_URL).netloc == parse.urlparse(field_dict['urlid']).netloc:
            # try slug field if it exists
            if slug_field in field_dict:
                kwargs = {slug_field: field_dict[slug_field]}
                sub_inst = field_model.objects.get(**kwargs)
            else:
                model, sub_inst = Model.resolve(field_dict['urlid'])
        # remote resource - get backlinked copy
        elif hasattr(field_model, 'urlid'):
            sub_inst = Model.get_or_create_external(field_model, field_dict['urlid'])
        return model, sub_inst

    def update_dict_value(self, attr, instance, value):
        info = model_meta.get_field_info(instance)
        relation_info = info.relations.get(attr)
        slug_field = Model.slug_field(relation_info.related_model)

        if slug_field in value:
            value = self.update_dict_value_when_id_is_provided(attr, instance, relation_info, slug_field, value)
        else:
            if 'urlid' in value:
                if parse.urlparse(settings.BASE_URL).netloc == parse.urlparse(value['urlid']).netloc:
                    model, oldObj = Model.resolve(value['urlid'])
                    value = self.update(instance=oldObj, validated_data=value)
                elif hasattr(relation_info.related_model, 'urlid'):
                    value = Model.get_or_create_external(relation_info.related_model, value['urlid'])
            else:
                value = self.update_dict_value_without_slug_field(attr, instance, relation_info, value)
        return value

    def update_dict_value_without_slug_field(self, attr, instance, relation_info, value):
        if relation_info.to_many:
            value = self.internal_create(validated_data=value, model=relation_info.related_model)
        else:
            rel = instance._meta.get_field(attr)
            reverse_attr_name = rel.remote_field.name
            many = rel.one_to_many or rel.many_to_many
            if many:
                value[reverse_attr_name] = [instance]
                oldObj = rel.model.object.get(id=value['urlid'])
            else:
                value[reverse_attr_name] = instance
                oldObj = getattr(instance, attr, None)

            if oldObj is None:
                value = self.internal_create(validated_data=value, model=relation_info.related_model)
            else:
                value = self.update(instance=oldObj, validated_data=value)
        return value

    def update_dict_value_when_id_is_provided(self, attr, instance, relation_info, slug_field, value):
        kwargs = {slug_field: value[slug_field]}
        if relation_info.to_many:
            manager = getattr(instance, attr)
            oldObj = manager._meta.model.objects.get(**kwargs)
        else:
            related_model = relation_info.related_model
            oldObj = related_model.objects.get(**kwargs)

        value = self.update(instance=oldObj, validated_data=value)
        return value

    def save_or_update_nested_list(self, instance, nested_fields):
        for (field_name, data) in nested_fields:
            manager = getattr(instance, field_name)
            field_model = manager.model
            slug_field = Model.slug_field(field_model)
            try:
                item_pk_to_keep = [obj_dict[slug_field] for obj_dict in data if slug_field in obj_dict]
            except TypeError:
                item_pk_to_keep = [getattr(obj, slug_field) for obj in data if hasattr(obj, slug_field)]

            if hasattr(manager, 'through'):
                manager.clear()
            else:
                manager.exclude(pk__in=item_pk_to_keep).delete()

            for item in data:
                if isinstance(item, Model):
                    item.save()
                    saved_item = item
                elif slug_field in item:
                    kwargs = {slug_field: item[slug_field]}
                    saved_item = self.get_or_create(field_model, item, kwargs)
                elif 'urlid' in item:
                    # has urlid and is a local resource
                    if not Model.is_external(item['urlid']):
                        model, old_obj = Model.resolve(item['urlid'])
                        if old_obj is not None:
                            saved_item = self.update(instance=old_obj, validated_data=item)
                        else:
                            saved_item = self.internal_create(validated_data=item, model=field_model)
                    # has urlid and is external resource
                    elif hasattr(field_model, 'urlid'):
                        kwargs = {'urlid': item['urlid']}
                        saved_item = self.get_or_create(field_model, item, kwargs)
                else:
                    rel = getattr(instance._meta.model, field_name).rel
                    try:
                        if rel.related_model == manager.model:
                            reverse_id = rel.remote_field.attname
                            item[reverse_id] = instance.pk
                    except AttributeError:
                        pass
                    saved_item = self.internal_create(validated_data=item, model=manager.model)

                if hasattr(manager, 'through') and manager.through._meta.auto_created:
                    #First remove to avoid duplicates
                    manager.remove(saved_item)
                    manager.add(saved_item)

    def get_or_create(self, field_model, item, kwargs):
        try:
            old_obj = field_model.objects.get(**kwargs)
            saved_item = self.update(instance=old_obj, validated_data=item)
        except field_model.DoesNotExist:
            saved_item = self.internal_create(validated_data=item, model=field_model)
        return saved_item


# Django imports
from django.apps import apps
from django.contrib.auth import get_user_model
from django.core.exceptions import FieldDoesNotExist
from django.shortcuts import get_object_or_404
from django.urls import include, path, re_path
from django.urls.resolvers import get_resolver
from django.utils.decorators import classonlymethod

# DjangoLDP imports
from djangoldp.filters import LocalObjectOnContainerPathBackend, SearchByQueryParamFilterBackend
from djangoldp.models import DynamicNestedField, LDPSource
from djangoldp.related import get_prefetch_fields
from djangoldp.utils import is_authenticated_user
from djangoldp.views.commons import JSONLDParser, JSONLDRenderer, NoCSRFAuthentication

# DRF imports
from rest_framework import status
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet

import logging

logger = logging.getLogger('djangoldp')
get_user_model()._meta.rdf_context = {"get_full_name": "rdfs:label"}


class LDPViewSetGenerator(ModelViewSet):
    """An extension of ModelViewSet that generates automatically URLs for the model"""
    model = None
    nested_fields = []
    model_prefix = None
    list_actions = {'get': 'list', 'post': 'create'}
    detail_actions = {'get': 'retrieve', 'put': 'update', 'patch': 'partial_update', 'delete': 'destroy'}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.lookup_field = LDPViewSetGenerator.get_lookup_arg(**kwargs)

    @classonlymethod
    def get_model(cls, **kwargs):
        '''gets the model in the arguments or in the viewset definition'''
        model = kwargs.get('model') or cls.model
        if isinstance(model, str):
            model = apps.get_model(model)
        return model

    @classonlymethod
    def get_lookup_arg(cls, **kwargs):
        return kwargs.get('lookup_url_kwarg') or cls.lookup_url_kwarg or kwargs.get('lookup_field') or \
               getattr(kwargs['model']._meta, 'lookup_field', 'pk') or cls.lookup_field

    @classonlymethod
    def get_detail_expr(cls, lookup_field=None, **kwargs):
        '''builds the detail url based on the lookup_field'''
        lookup_field = lookup_field or cls.get_lookup_arg(**kwargs)
        lookup_group = r'\d' if lookup_field == 'pk' else r'[\w\-\.]'
        return r'(?P<{}>{}+)/'.format(lookup_field, lookup_group)

    @classonlymethod
    def build_nested_view_set(cls, view_set=None):
        '''returns the the view_set parameter mixed into the LDPNestedViewSet class'''
        if view_set is not None:
            class LDPNestedCustomViewSet(LDPNestedViewSet, view_set):
                pass
            return LDPNestedCustomViewSet
        return LDPNestedViewSet

    @classonlymethod
    def urls(cls, **kwargs):
        '''constructs urls list for model passed in kwargs'''
        kwargs['model'] = cls.get_model(**kwargs)
        model_name = kwargs['model']._meta.object_name.lower()
        if kwargs.get('model_prefix'):
            model_name = '{}-{}'.format(kwargs['model_prefix'], model_name)
        detail_expr = cls.get_detail_expr(**kwargs)
        # Gets permissions on the model if not explicitely passed to the view
        if not 'permission_classes' in kwargs and hasattr(kwargs['model']._meta, 'permission_classes'):
            kwargs['permission_classes'] = kwargs['model']._meta.permission_classes

        urls = [
            path('', cls.as_view(cls.list_actions, **kwargs), name='{}-list'.format(model_name)),
            re_path('^' + detail_expr + '$', cls.as_view(cls.detail_actions, **kwargs),
                    name='{}-detail'.format(model_name)),
        ]

        # append nested fields to the urls list
        for field_name in kwargs.get('nested_fields') or cls.nested_fields:
            try:
                nested_field = kwargs['model']._meta.get_field(field_name)
                nested_model = nested_field.related_model
                field_name_to_parent = nested_field.remote_field.name
            except FieldDoesNotExist:
                nested_model = getattr(kwargs['model'], field_name).field.model
                nested_field = getattr(kwargs['model'], field_name).field.remote_field
                field_name_to_parent = getattr(kwargs['model'], field_name).field.name

            # urls should be called from _nested_ view set, which may need a custom view set mixed in
            view_set = getattr(nested_model._meta, 'view_set', None)
            nested_view_set = cls.build_nested_view_set(view_set)

            urls.append(re_path('^' + detail_expr + field_name + '/',
                    nested_view_set.urls(
                    model=nested_model,
                    model_prefix=kwargs['model']._meta.object_name.lower(), # prefix with parent name
                    lookup_field=getattr(nested_model._meta, 'lookup_field', 'pk'),
                    exclude=(field_name_to_parent,) if nested_field.one_to_many else (),
                    permission_classes=getattr(nested_model._meta, 'permission_classes', []),
                    nested_field_name=field_name,
                    fields=getattr(nested_model._meta, 'serializer_fields', []),
                    nested_fields=[],
                    parent_model=kwargs['model'],
                    parent_lookup_field=cls.get_lookup_arg(**kwargs),
                    nested_field=nested_field,
                    field_name_to_parent=field_name_to_parent)))

        return include(urls)



# LDPViewSetGenerator is a ModelViewSet (DRF) with methods to automatically generate model urls
class LDPViewSet(LDPViewSetGenerator):
    """An automatically generated viewset that serves models following the Linked Data Platform convention"""
    fields = None
    exclude = None
    renderer_classes = (JSONLDRenderer,)
    parser_classes = (JSONLDParser,)
    authentication_classes = (NoCSRFAuthentication,)
    filter_backends = [SearchByQueryParamFilterBackend, LocalObjectOnContainerPathBackend]
    prefetch_fields = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # attach filter backends based on permissions classes, to reduce the queryset based on these permissions
        # https://www.django-rest-framework.org/api-guide/filtering/#generic-filtering
        self.filter_backends = type(self).filter_backends + list({perm_class().get_filter_backend(self.model)
                for perm_class in self.permission_classes if hasattr(perm_class(), 'get_filter_backend')})
        if None in self.filter_backends:
            self.filter_backends.remove(None)
    
    def filter_queryset(self, queryset):
        if self.request.user.is_superuser:
            return queryset
        return super().filter_queryset(queryset)

    def check_permissions(self, request):
        if request.user.is_superuser:
            return True
        return super().check_permissions(request)

    def check_object_permissions(self, request, obj):
        if request.user.is_superuser:
            return True
        return super().check_object_permissions(request, obj)
    
    def get_depth(self) -> int:
        if getattr(self, 'force_depth', None):
            #TODO: this exception on depth for writing should be handled by the serializer itself
            return self.force_depth
        if hasattr(self, 'request') and 'HTTP_DEPTH' in self.request.META:
            return int(self.request.META['HTTP_DEPTH'])
        if hasattr(self, 'depth'):
            return self.depth
        return getattr(self.model._meta, 'depth', 0)

    def get_serializer_class(self):
        model_name = self.model._meta.object_name.lower()
        try:
            lookup_field = get_resolver().reverse_dict[model_name + '-detail'][0][0][1][0]
        except:
            lookup_field = 'urlid'
        
        meta_args = {'model': self.model, 'extra_kwargs': {
                '@id': {'lookup_field': lookup_field}},
                'depth': self.get_depth(),
                'extra_fields': self.nested_fields}

        if self.fields:
            meta_args['fields'] = self.fields
        else:
            meta_args['exclude'] = self.exclude or getattr(self.model._meta, 'serializer_fields_exclude', ())
        # create the Meta class to associate to LDPSerializer, using meta_args param

        from djangoldp.serializers import LDPSerializer
        if self.serializer_class is None:
            self.serializer_class = LDPSerializer

        parent_meta = (self.serializer_class.Meta,) if hasattr(self.serializer_class, 'Meta') else ()
        meta_class = type('Meta', parent_meta, meta_args)

        return type(self.serializer_class)(self.model._meta.object_name.lower() + 'Serializer',
                                   (self.serializer_class,),
                                   {'Meta': meta_class})

    # The chaining of filter through | may lead to duplicates and distinct should only be applied in the end.
    def filter_queryset(self, queryset):
        return super().filter_queryset(queryset).distinct()

    def create(self, request, *args, **kwargs):
        self.force_depth = 10
        serializer = self.get_serializer(data=request.data)
        self.force_depth = None
        serializer.is_valid(raise_exception=True)

        self.perform_create(serializer)
        response_serializer = self.get_serializer()
        data = response_serializer.to_representation(serializer.instance)
        headers = self.get_success_headers(data)
        return Response(data, status=status.HTTP_201_CREATED, headers=headers)

    def update(self, request, *args, **kwargs):
        partial = kwargs.pop('partial', False)
        instance = self.get_object()
        self.force_depth = 10
        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        self.force_depth = None
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)

        if getattr(instance, '_prefetched_objects_cache', None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        response_serializer = self.get_serializer()
        data = response_serializer.to_representation(serializer.instance)
        return Response(data)

    def perform_create(self, serializer, **kwargs):
        if hasattr(self.model._meta, 'auto_author') and isinstance(self.request.user, get_user_model()):
            kwargs[self.model._meta.auto_author] = get_user_model().objects.get(pk=self.request.user.pk)
        return serializer.save(**kwargs)

    def get_queryset(self, *args, **kwargs):
        if self.model:
            queryset = self.model.objects.all()
        else:
            queryset = super(LDPViewSet, self).get_queryset(*args, **kwargs)
        if self.prefetch_fields is None:
            self.prefetch_fields = get_prefetch_fields(self.model, self.get_serializer(), self.get_depth())
        return queryset.prefetch_related(*self.prefetch_fields)

    def dispatch(self, request, *args, **kwargs):
        '''overriden dispatch method to append some custom headers'''
        response = super(LDPViewSet, self).dispatch(request, *args, **kwargs)
        response["Accept-Post"] = "application/ld+json"

        if response.status_code in [201, 200] and '@id' in response.data:
            response["Location"] = str(response.data['@id'])
        else:
            pass

        if is_authenticated_user(request.user):
            try:
                response['User'] = request.user.urlid
            except AttributeError:
                pass
        return response


class LDPNestedViewSet(LDPViewSet):
    """
    A special case of LDPViewSet serving objects of a relation of a given object
    (e.g. members of a group, or skills of a user)
    """
    parent_model = None
    parent_lookup_field = None
    nested_field = None
    nested_field_name = None
    field_name_to_parent = None

    def get_parent(self):
        return get_object_or_404(self.parent_model, **{self.parent_lookup_field: self.kwargs[self.parent_lookup_field]})

    def perform_create(self, serializer, **kwargs):
        kwargs[self.field_name_to_parent] = self.get_parent()
        super().perform_create(serializer, **kwargs)

    def get_queryset(self, *args, **kwargs):
        related = getattr(self.get_parent(), self.nested_field_name)
        if self.nested_field.many_to_many or self.nested_field.one_to_many:
            if isinstance(self.nested_field, DynamicNestedField):
                return related()
            return related.all()
        if self.nested_field.one_to_one or self.nested_field.many_to_one:
            return type(related).objects.filter(pk=related.pk)


        return response


class LDPSourceViewSet(LDPViewSet):
    model = LDPSource
    federation = None

    def get_queryset(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs).filter(federation=self.kwargs['federation'])



from django.conf import settings
from django.utils.http import url_has_allowed_host_and_scheme
from django.shortcuts import redirect
from djangoldp.models import Model
from django.http import HttpResponse

class AllowOnlySiteUrl:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        if(url_has_allowed_host_and_scheme(request.get_raw_uri(), allowed_hosts=settings.SITE_URL) or response.status_code != 200):
            return response
        else:
            return redirect('{}{}'.format(settings.SITE_URL, request.path), permanent=True)


class AllowRequestedCORSMiddleware:
    '''A CORS Middleware which allows the domains requested by the request'''
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if request.method == 'OPTIONS':
            # Return an empty 200 response for OPTIONS requests
            # The CORS headers will be added by AllowRequestedCORSMiddleware later
            response = HttpResponse(status=200)
        else:
            response = self.get_response(request)

        response["Access-Control-Allow-Origin"] = request.headers.get('origin')
        response["Access-Control-Allow-Methods"] = "GET,POST,PUT,PATCH,DELETE,OPTIONS,HEAD"
        response["Access-Control-Allow-Headers"] = \
            getattr(settings, 'OIDC_ACCESS_CONTROL_ALLOW_HEADERS',
                    "authorization, Content-Type, if-match, accept, DPoP, cache-control, prefer")
        response["Access-Control-Expose-Headers"] = "Location, User"
        response["Access-Control-Allow-Credentials"] = 'true'

        return response


class OptionsResponseMiddleware:
    """Middleware that returns a 200 response for OPTIONS requests.
    
    This should be placed BEFORE the AllowRequestedCORSMiddleware in the stack
    to ensure OPTIONS requests receive proper CORS headers.
    
    This middleware is not enabled by default. To use it, add
    'djangoldp.middleware.OptionsResponseMiddleware' to your MIDDLEWARE setting
    before 'djangoldp.middleware.AllowRequestedCORSMiddleware'.
    """
    
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if request.method == 'OPTIONS':
            # Return an empty 200 response for OPTIONS requests
            # The CORS headers will be added by AllowRequestedCORSMiddleware later
            return HttpResponse(status=200)
        
        # For all other requests, proceed as normal
        return self.get_response(request)


import os
import json
import requests
from django.core.management.base import BaseCommand
from django.conf import settings
from django.apps import apps
from urllib.parse import urlparse, urljoin

class StaticContentGenerator:
    def __init__(self, stdout, style):
        self.stdout = stdout
        self.style = style
        self.base_uri = getattr(settings, 'BASE_URL', '')
        self.max_depth = getattr(settings, 'MAX_RECURSION_DEPTH', 5)
        self.request_timeout = getattr(settings, 'SSR_REQUEST_TIMEOUT', 10)
        self.regenerated_urls = set()
        self.failed_urls = set()
        self.output_dir = 'ssr'
        self.output_dir_filtered = 'ssr_filtered'

    def generate_content(self):
        self._create_output_directory()
        for model in self._get_static_models():
            self._process_model(model)

    def _create_output_directory(self):
        os.makedirs(self.output_dir, exist_ok=True)
        os.makedirs(self.output_dir_filtered, exist_ok=True)

    def _get_static_models(self):
        return [model for model in apps.get_models() if hasattr(model._meta, 'static_version')]

    def _process_model(self, model):
        self.stdout.write(f"Generating content for model: {model}")
        url = self._build_url(model)
        if url not in self.regenerated_urls and url not in self.failed_urls:
            self._fetch_and_save_content(model, url, self.output_dir)
        else:
            self.stdout.write(self.style.WARNING(f'Skipping {url} as it has already been fetched'))
        if hasattr(model._meta, 'static_params'):
            url = self._build_url(model, True)
            if url not in self.regenerated_urls and url not in self.failed_urls:
                self._fetch_and_save_content(model, url, self.output_dir_filtered)
            else:
                self.stdout.write(self.style.WARNING(f'Skipping {url} as it has already been fetched'))

    def _build_url(self, model, use_static_params=False):
        container_path = model.get_container_path()
        url = urljoin(self.base_uri, container_path)
        if hasattr(model._meta, 'static_params') and use_static_params:
            url += '?' + '&'.join(f'{k}={v}' for k, v in model._meta.static_params.items())
        return url

    def _fetch_and_save_content(self, model, url, output_dir):
        try:
            response = requests.get(url, timeout=self.request_timeout)
            if response.status_code == 200:
                content = self._update_ids_and_fetch_associated(response.text)
                self._save_content(model, url, content, output_dir)
            else:
                self.stdout.write(self.style.ERROR(f'Failed to fetch content from {url}: HTTP {response.status_code}'))
        except requests.exceptions.RequestException as e:
            self.stdout.write(self.style.ERROR(f'Error fetching content from {url}: {str(e)}'))

    def _save_content(self, model, url, content, output_dir):
        relative_path = urlparse(url).path.strip('/')
        file_path = os.path.join(output_dir, relative_path)
        if file_path.endswith('/'):
            file_path = file_path[:-1]
        if not file_path.endswith('.jsonld'):
            file_path += '.jsonld'
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        try:
            with open(file_path, 'w', encoding='utf-8') as f:
                f.write(content)
            self.stdout.write(self.style.SUCCESS(f'Successfully saved content for {model._meta.model_name} from {url} to {file_path}'))
        except IOError as e:
            self.stdout.write(self.style.ERROR(f'Error saving content for {model._meta.model_name}: {str(e)}'))

    def _update_ids_and_fetch_associated(self, content, depth=0):
        if depth > self.max_depth:
            return content

        try:
            data = json.loads(content)
            self._process_data(data, depth)
            return json.dumps(data)
        except json.JSONDecodeError as e:
            self.stdout.write(self.style.ERROR(f'Failed to decode JSON: {e}'))
            return content

    def _process_data(self, data, depth):
        if isinstance(data, dict):
            self._process_item(data, depth)
        elif isinstance(data, list):
            for item in data:
                if isinstance(item, dict):
                    self._process_item(item, depth)

    def _process_item(self, item, depth):
        if '@id' in item:
            self._update_and_fetch_content(item, depth)

        for value in item.values():
            if isinstance(value, (dict, list)):
                self._process_data(value, depth)

    def _update_and_fetch_content(self, item, depth):
        original_id = item['@id']
        parsed_url = urlparse(original_id)

        if not parsed_url.netloc:
            original_id = urljoin(self.base_uri, original_id)
            parsed_url = urlparse(original_id)

        path = parsed_url.path
        if path.startswith(urlparse(self.base_uri).path):
            path = path[len(urlparse(self.base_uri).path):]

        new_id = f'/ssr{path}'
        item['@id'] = urljoin(self.base_uri, new_id)

        self._fetch_and_save_associated_content(original_id, path, depth)

    def _rewrite_ids_before_saving(self, data):
        if isinstance(data, dict):
            if '@id' in data:
                original_id = data['@id']
                parsed_url = urlparse(data['@id'])
                if not parsed_url.netloc:
                    content_id = urljoin(self.base_uri, original_id)
                    parsed_url = urlparse(content_id)

                if 'ssr/' not in data['@id']:
                    path = parsed_url.path
                    if path.startswith(urlparse(self.base_uri).path):
                        path = path[len(urlparse(self.base_uri).path):]

                    new_id = f'/ssr{path}'
                    data['@id'] = urljoin(self.base_uri, new_id)
            for value in data.values():
                if isinstance(value, (dict, list)):
                    self._rewrite_ids_before_saving(value)
        elif isinstance(data, list):
            for item in data:
                self._rewrite_ids_before_saving(item)
        return data

    def _fetch_and_save_associated_content(self, url, new_path, depth):
        if url in self.regenerated_urls:
            self.stdout.write(self.style.WARNING(f'Skipping {url} as it has already been fetched'))
            return
        if url in self.failed_urls:
            self.stdout.write(self.style.WARNING(f'Skipping {url} as it has already been tried and failed'))
            return

        file_path = os.path.join(self.output_dir, new_path.strip('/'))
        if file_path.endswith('/'):
            file_path = file_path[:-1]
        if not file_path.endswith('.jsonld'):
            file_path += '.jsonld'
        os.makedirs(os.path.dirname(file_path), exist_ok=True)

        try:
            response = requests.get(url, timeout=self.request_timeout)
            if response.status_code == 200:
                updated_content = json.loads(self._update_ids_and_fetch_associated(response.text, depth + 1))
                updated_content = self._rewrite_ids_before_saving(updated_content)

                with open(file_path, 'w', encoding='utf-8') as f:
                    f.write(json.dumps(updated_content))
                self.regenerated_urls.add(url)
                self.stdout.write(self.style.SUCCESS(f'Successfully fetched and saved associated content from {url} to {file_path}'))
            else:
                self.failed_urls.add(url)
                self.stdout.write(self.style.ERROR(f'Failed to fetch associated content from {url}: HTTP {response.status_code}'))
        except requests.exceptions.RequestException as e:
            self.stdout.write(self.style.ERROR(f'Error fetching associated content from {url}: {str(e)}'))
        except IOError as e:
            self.stdout.write(self.style.ERROR(f'Error saving associated content from {url}: {str(e)}'))

class Command(BaseCommand):
    help = 'Generate static content for models having the static_version meta attribute set to 1/true'

    def handle(self, *args, **options):
        generator = StaticContentGenerator(self.stdout, self.style)
        generator.generate_content()


from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings
from rest_framework.test import APIClient, APIRequestFactory
from rest_framework.utils import json

from djangoldp.tests.models import Circle, Conversation, Project


class TestCache(TestCase):

    def setUp(self):
        self.factory = APIRequestFactory()
        self.client = APIClient()
        self.user = get_user_model().objects.create_user(username='john', email='jlennon@beatles.com',
                                                         password='glass onion', first_name='John')
        self.client.force_authenticate(self.user)

    def tearDown(self):
        setattr(Circle._meta, 'depth', 0)
        setattr(Circle._meta, 'empty_containers', [])

    # test container cache after new resource added
    @override_settings(SERIALIZER_CACHE=True)
    def test_save_fk_graph_with_nested(self):
        response = self.client.get('/batchs/', content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)

        post = {
            '@graph': [
                {
                    'https://cdn.startinblox.com/owl#title': "title",
                    'https://cdn.startinblox.com/owl#invoice': {
                        '@id': "_.123"
                    }
                },
                {
                    '@id': "_.123",
                    'https://cdn.startinblox.com/owl#title': "title 2"
                }
            ]
        }

        response = self.client.post('/batchs/', data=json.dumps(post), content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)

        response = self.client.get('/batchs/', content_type='application/ld+json')
        self.assertIn('ldp:contains', response.data)
        self.assertEqual(response.data['ldp:contains'][0]['title'], "title")
        self.assertEqual(response.data['ldp:contains'][0]['invoice']['title'], "title 2")

    # test resource cache after it is updated
    @override_settings(SERIALIZER_CACHE=True)
    def test_update_with_new_fk_relation(self):
        conversation = Conversation.objects.create(author_user=self.user, description="conversation description")
        response = self.client.get('/conversations/{}/'.format(conversation.pk), content_type='application/ld+json')
        body = [
            {
                '@id': "/conversations/{}/".format(conversation.pk),
                'https://cdn.startinblox.com/owl#description': "conversation update",
                'https://cdn.startinblox.com/owl#peer_user': {
                    '@id': self.user.urlid,
                }
            }
        ]
        response = self.client.put('/conversations/{}/'.format(conversation.pk), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)

        response = self.client.get('/conversations/{}/'.format(conversation.pk), content_type='application/ld+json')
        self.assertIn('peer_user', response.data)
        self.assertEqual('conversation update', response.data['description'])
        self.assertEqual(response.data['peer_user']['@id'], self.user.urlid)
        self.assertIn('@type', response.data['peer_user'])

    # test resource cache after it is updated - external resource
    @override_settings(SERIALIZER_CACHE=True)
    def test_update_with_new_fk_relation_external(self):
        conversation = Conversation.objects.create(author_user=self.user, description="conversation description")
        response = self.client.get('/conversations/{}/'.format(conversation.pk), content_type='application/ld+json')
        external_user = get_user_model().objects.create_user(username='external', email='jlennon@beatles.com',
                                                             password='glass onion', urlid='https://external.com/users/external/')
        body = [
            {
                '@id': "/conversations/{}/".format(conversation.pk),
                'https://cdn.startinblox.com/owl#description': "conversation update",
                'https://cdn.startinblox.com/owl#peer_user': {
                    '@id': external_user.urlid,
                }
            }
        ]
        response = self.client.put('/conversations/{}/'.format(conversation.pk), data=json.dumps(body),
                                   content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)

        response = self.client.get('/conversations/{}/'.format(conversation.pk), content_type='application/ld+json')
        self.assertIn('peer_user', response.data)
        self.assertEqual('conversation update', response.data['description'])
        self.assertIn('@id', response.data['peer_user'])
        # serialize external id and only external id
        self.assertEqual(response.data['peer_user']['@id'], external_user.urlid)
        self.assertIn('@type', response.data['peer_user'])
        self.assertEqual(len(response.data['peer_user']), 2)

    # test container cache after member is deleted by view
    @override_settings(SERIALIZER_CACHE=True)
    def test_cached_container_deleted_resource_view(self):
        conversation = Conversation.objects.create(author_user=self.user, description="conversation description")
        response = self.client.get('/conversations/', content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 1)

        response = self.client.delete('/conversations/{}/'.format(conversation.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 204)

        response = self.client.get('/conversations/', content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 0)

    # test container cache after member is deleted manually
    @override_settings(SERIALIZER_CACHE=True)
    def test_cached_container_deleted_resource_manual(self):
        conversation = Conversation.objects.create(author_user=self.user, description="conversation description")
        response = self.client.get('/conversations/', content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 1)

        conversation.delete()

        response = self.client.get('/conversations/', content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 0)

    # test resource cache after it is deleted manually
    @override_settings(SERIALIZER_CACHE=True)
    def test_cached_resource_deleted_resource_manual(self):
        conversation = Conversation.objects.create(author_user=self.user, description="conversation description")
        response = self.client.get('/conversations/{}/'.format(conversation.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)

        conversation.delete()

        response = self.client.get('/conversations/{}/'.format(conversation.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 404)

    # test container cache following m2m_changed - Project (which inherits from djangoldp.models.Model)
    @override_settings(SERIALIZER_CACHE=True)
    def test_cached_container_m2m_changed_project(self):
        project = Project.objects.create(description='Test')
        response = self.client.get('/projects/{}/members/'.format(project.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 0)

        project.members.add(self.user)
        response = self.client.get('/projects/{}/members/'.format(project.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 1)

        project.members.remove(self.user)
        response = self.client.get('/projects/{}/members/'.format(project.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 0)

        project.members.add(self.user)
        project.members.clear()
        response = self.client.get('/projects/{}/members/'.format(project.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 0)

    # test container cache following m2m_changed - Conversation (which does not inherit from djangoldp.models.Model)
    @override_settings(SERIALIZER_CACHE=True)
    def test_cached_container_m2m_changed_conversation(self):
        conversation = Conversation.objects.create(author_user=self.user, description="conversation description")
        response = self.client.get('/conversations/{}/observers/'.format(conversation.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 0)

        conversation.observers.add(self.user)
        response = self.client.get('/conversations/{}/observers/'.format(conversation.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 1)

        conversation.observers.remove(self.user)
        response = self.client.get('/conversations/{}/observers/'.format(conversation.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 0)

        conversation.observers.add(self.user)
        conversation.observers.clear()
        response = self.client.get('/conversations/{}/observers/'.format(conversation.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['ldp:contains']), 0)

    # test cache working inside of the nested field (serializer) of another object
    @override_settings(SERIALIZER_CACHE=True)
    def test_cached_container_serializer_nested_field(self):
        project = Project.objects.create(description='Test')
        response = self.client.get('/projects/{}/'.format(project.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['members']['ldp:contains']), 0)

        project.members.add(self.user)
        response = self.client.get('/projects/{}/'.format(project.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['members']['ldp:contains']), 1)

        project.members.remove(self.user)
        response = self.client.get('/projects/{}/'.format(project.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['members']['ldp:contains']), 0)

        project.members.add(self.user)
        project.members.clear()
        response = self.client.get('/projects/{}/'.format(project.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['members']['ldp:contains']), 0)

    # test cache working on a serialized nested field at higher depth
    @override_settings(SERIALIZER_CACHE=True)
    def test_cache_depth_2(self):
        setattr(Circle._meta, 'depth', 2)

        circle = Circle.objects.create(description='Test')
        response = self.client.get('/circles/{}/'.format(circle.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['members']['user_set']), 0)

        circle.members.user_set.add(self.user)
        response = self.client.get('/circles/{}/'.format(circle.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['members']['user_set']), 1)
        # assert the depth is applied
        self.assertIn('first_name', response.data['members']['user_set'][0])
        self.assertEqual(response.data['members']['user_set'][0]['first_name'], self.user.first_name)

        # make a change to the _user_
        self.user.first_name = "Alan"
        self.user.save()

        # assert that the use under the circles members has been updated
        response = self.client.get('/circles/{}/'.format(circle.pk), content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data['members']['user_set']), 1)
        self.assertIn('first_name', response.data['members']['user_set'][0])
        self.assertEqual(response.data['members']['user_set'][0]['first_name'], self.user.first_name)

    # test the cache behaviour when empty_containers is an active setting
    @override_settings(SERIALIZER_CACHE=True)
    def test_cache_empty_container(self):
        setattr(Circle._meta, 'depth', 1)
        setattr(Circle._meta, 'empty_containers', ['members'])

        circle = Circle.objects.create(name='test', description='test', owner=self.user)
        circle.members.user_set.add(self.user)

        # make one call on the parent
        response = self.client.get('/circles/', content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('ldp:contains', response.data)
        self.assertEqual(len(response.data['ldp:contains']), 1)
        self.assertIn('members', response.data['ldp:contains'][0])
        self.assertIn('@id', response.data['ldp:contains'][0]['members'])
        self.assertIn('user_set', response.data['ldp:contains'][0]['members'])
        self.assertEqual(len(response.data['ldp:contains'][0]['members']['user_set']), 1)
        self.assertIn('@id', response.data['ldp:contains'][0]['members']['user_set'][0])
        self.assertEqual(response.data['ldp:contains'][0]['members']['user_set'][0]['@type'], 'foaf:user')

        # and a second on the child
        response = self.client.get(response.data['ldp:contains'][0]['members']['@id'], content_type='application/ld+json')
        self.assertEqual(response.status_code, 200)
        self.assertIn('@id', response.data)
        self.assertEqual(len(response.data['user_set']), 1)



import json
import logging
import os
import time

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.http import HttpResponseNotFound, JsonResponse
from django.urls.resolvers import get_resolver

from rest_framework.renderers import JSONRenderer

logger = logging.getLogger('djangoldp')


def serve_static_content(request, path):

    if request.method != "GET":
        resolver = get_resolver()
        match = resolver.resolve("/" + path)
        request.user = AnonymousUser()
        return match.func(request, *match.args, **match.kwargs)

    server_url = getattr(settings, "BASE_URL", "http://localhost")

    is_filtered = request.GET.get('search-fields', False)

    output_dir = "ssr"
    output_dir_filtered = "ssr_filtered"
    if not os.path.exists(output_dir):
        os.makedirs(output_dir, exist_ok=True)
    if not os.path.exists(output_dir_filtered):
        os.makedirs(output_dir_filtered, exist_ok=True)

    file_path = os.path.join(output_dir if not is_filtered else output_dir_filtered, path[:-1])
    if not file_path.endswith(".jsonld"):
        file_path += ".jsonld"

    if os.path.exists(file_path):
        current_time = time.time()
        file_mod_time = os.path.getmtime(file_path)
        time_difference = current_time - file_mod_time
        if time_difference > 24 * 60 * 60:
            os.remove(file_path)

    if not os.path.exists(file_path):

        resolver = get_resolver()
        match = resolver.resolve("/" + path)
        request.user = AnonymousUser()
        response = match.func(request, *match.args, **match.kwargs)
        if response.status_code == 200:
            directory = os.path.dirname(file_path)
            if not os.path.exists(directory):
                os.makedirs(directory)
            json_content = JSONRenderer().render(response.data)
            with open(file_path, "w", encoding="utf-8") as f:
                f.write(
                    json_content.decode("utf-8")
                    .replace('"@id":"' + server_url, '"@id":"' + server_url + "/ssr")
                    .replace(
                        '"@id":"' + server_url + "/ssr/ssr",
                        '"@id":"' + server_url + "/ssr",
                    )[:-1]
                    + ',"@context": "'
                    + getattr(
                        settings,
                        "LDP_RDF_CONTEXT",
                        "https://cdn.startinblox.com/owl/context.jsonld",
                    )
                    + '"}'
                )

    if os.path.exists(file_path):
        with open(file_path, "r", encoding="utf-8") as file:
            content = file.read()

        try:
            json_content = json.loads(content)
            return JsonResponse(
                json_content,
                safe=False,
                status=200,
                content_type="application/ld+json",
                headers={
                    "Access-Control-Allow-Origin": "*",
                    "Cache-Control": "public, max-age=3600",
                },
            )
        except json.JSONDecodeError:
            pass

    return HttpResponseNotFound("File not found")



import json
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from guardian.models import GroupObjectPermission
from rest_framework.test import APIRequestFactory, APIClient, APITestCase
from djangoldp.tests.models import AnonymousReadOnlyPost, AuthenticatedOnlyPost, ReadOnlyPost, DoubleInheritModel, \
    ReadAndCreatePost, OwnedResource, RestrictedCircle, RestrictedResource, ANDPermissionsDummy, ORPermissionsDummy

class TestPermissions(APITestCase):
    def setUp(self):
        self.factory = APIRequestFactory()
        self.client = APIClient()

    def authenticate(self):
        self.user = get_user_model().objects.create_user(username='random', email='random@user.com', password='Imrandom')
        self.client = APIClient(enforce_csrf_checks=True)
        self.client.force_authenticate(user=self.user)

    def check_can_add(self, url, status_code=201, field='content', extra_content={}):
        data = extra_content
        extra_content[f"https://cdn.startinblox.com/owl#{field}"] = "new post"
        response = self.client.post(url, data=json.dumps(data), content_type='application/ld+json')
        self.assertEqual(response.status_code, status_code)
        if status_code == 201:
            self.assertIn('@id', response.data)
            return response.data['@id']
    
    def check_can_change(self, id, status_code=200, field='content'):
        data = { f"https://cdn.startinblox.com/owl#{field}": "changed post" }
        response = self.client.put(id, data=json.dumps(data), content_type='application/ld+json')
        self.assertEqual(response.status_code, status_code)
        if status_code == 200:
            self.assertIn('@id', response.data)
            self.assertEqual(response.data['@id'], id)
        
    def check_can_view_one(self, id, status_code=200):
        response = self.client.get(id, content_type='application/ld+json')
        self.assertEqual(response.status_code, status_code)
        if status_code == 200:
            self.assertEqual(response.data['@id'], id)

    def check_can_view(self, url, ids, status_code=200):
        response = self.client.get(url, content_type='application/ld+json')
        self.assertEqual(response.status_code, status_code)
        if status_code == 200:
            self.assertEqual(len(response.data['ldp:contains']), len(ids))
            for resource, id in zip(response.data['ldp:contains'], ids):
                self.assertEqual(resource['@id'], id)
        for id in ids:
            self.check_can_view_one(id, status_code)
        

    def test_permissionless_model(self):
        id = self.check_can_add('/posts/')
        self.check_can_view('/posts/', [id])

    def test_anonymous_readonly(self):
        post = AnonymousReadOnlyPost.objects.create(content = "test post")
        self.check_can_view('/anonymousreadonlyposts/', [post.urlid])
        self.check_can_add('/anonymousreadonlyposts/', 403)
        self.check_can_change(post.urlid, 403)

        self.authenticate()
        self.check_can_add('/anonymousreadonlyposts/')
        self.check_can_change(post.urlid)
    
    def test_authenticated_only(self):
        post = AuthenticatedOnlyPost.objects.create(content = "test post")
        self.check_can_view('/authenticatedonlyposts/', [post.urlid], 403)
        self.check_can_add('/authenticatedonlyposts/', 403)
        self.check_can_change(post.urlid, 403)
        post.delete()

        self.authenticate()
        #When authenticated it should behave like a non protected model
        id = self.check_can_add('/authenticatedonlyposts/')
        self.check_can_view('/authenticatedonlyposts/', [id])
        self.check_can_change(id)

    def test_readonly(self):
        post = ReadOnlyPost.objects.create(content = "test post")
        self.check_can_view('/readonlyposts/', [post.urlid])
        self.check_can_add('/readonlyposts/', 403)
        self.check_can_change(post.urlid, 403)

    def test_readandcreate(self):
        post = ReadAndCreatePost.objects.create(content = "test post")
        self.check_can_view('/readandcreateposts/', [post.urlid])
        self.check_can_add('/readandcreateposts/')
        self.check_can_change(post.urlid, 403)
        
    def test_owner_permissions(self):
        self.authenticate()
        them = get_user_model().objects.create_user(username='them', email='them@user.com', password='itstheirsecret')
        mine = OwnedResource.objects.create(description="Mine!", user=self.user)
        theirs = OwnedResource.objects.create(description="Theirs", user=them)
        noones = OwnedResource.objects.create(description="I belong to NO ONE!")
        self.check_can_view('/ownedresources/', [mine.urlid]) #checks I can access mine and only mine
        self.check_can_change(mine.urlid)
        self.check_can_view_one(theirs.urlid, 404)
        self.check_can_change(theirs.urlid, 404)
        self.check_can_view_one(noones.urlid, 404)
        self.check_can_change(noones.urlid, 404)


    def check_permissions(self, obj, group, required_perms):
        perms = GroupObjectPermission.objects.filter(group=group)
        for perm in perms:
            self.assertEqual(perm.content_type.model, obj._meta.model_name)
            self.assertEqual(perm.object_pk, str(obj.pk))
        self.assertEqual(set(perms.values_list('permission__codename', flat=True)),
                         {f'{perm}_{obj._meta.model_name}' for perm in required_perms})
    
    def create_circles(self):
        self.authenticate()
        self.user.user_permissions.add(Permission.objects.get(codename='view_restrictedcircle'))
        them = get_user_model().objects.create_user(username='them', email='them@user.com', password='itstheirsecret')
        mine = RestrictedCircle.objects.create(name="mine", description="Mine!", owner=self.user)
        theirs = RestrictedCircle.objects.create(name="theirs", description="Theirs", owner=them)
        noones = RestrictedCircle.objects.create(name="no one's", description="I belong to NO ONE!")
        return mine, theirs, noones

    def test_role_permissions(self):
        mine, theirs, noones = self.create_circles()
        self.assertIn(self.user, mine.members.user_set.all())
        self.assertIn(self.user, mine.admins.user_set.all())
        self.assertNotIn(self.user, theirs.members.user_set.all())
        self.assertNotIn(self.user, theirs.admins.user_set.all())
        self.assertNotIn(self.user, noones.members.user_set.all())
        self.assertNotIn(self.user, noones.admins.user_set.all())

        self.check_can_view('/restrictedcircles/', [mine.urlid]) #check filtering

        self.check_permissions(mine, mine.members, RestrictedCircle._meta.permission_roles['members']['perms'])
        self.check_permissions(mine, mine.admins, RestrictedCircle._meta.permission_roles['admins']['perms'])

    def test_inherit_permissions(self):
        mine, theirs, noones = self.create_circles()
        myresource = RestrictedResource.objects.create(content="mine", circle=mine)
        their_resource = RestrictedResource.objects.create(content="theirs", circle=theirs)
        noones_resource  = RestrictedResource.objects.create(content="noones", circle=noones)

        self.check_can_view('/restrictedresources/', [myresource.urlid])
        self.check_can_change(myresource.urlid)
        self.check_can_change(their_resource.urlid, 404)
        self.check_can_change(noones_resource.urlid, 404)

    def test_inherit_several_permissions(self):
        mine, theirs, noones = self.create_circles()
        ro_resource = ReadOnlyPost.objects.create(content="read only")
        myresource = DoubleInheritModel.objects.create(content="mine", circle=mine, ro_ancestor=None)
        some = DoubleInheritModel.objects.create(content="some", circle=theirs, ro_ancestor=ro_resource)
        other = DoubleInheritModel.objects.create(content="other", circle=noones, ro_ancestor=None)

        self.check_can_view('/doubleinheritmodels/', [myresource.urlid, some.urlid])
        self.check_can_change(myresource.urlid)
        self.check_can_change(some.urlid, 403)
        self.check_can_change(other.urlid, 404)

    def test_inherit_permissions_none(self):
        id = self.check_can_add('/doubleinheritmodels/')
        resource = DoubleInheritModel.objects.get(urlid=id)
        self.check_can_view('/doubleinheritmodels/', [resource.urlid])

        circle = RestrictedCircle.objects.create()
        id = self.check_can_add('/doubleinheritmodels/', 404, extra_content={'https://cdn.startinblox.com/owl#circle': {'@id': circle.urlid}})

    
    def test_and_permissions(self):
        self.authenticate()
        abc = ANDPermissionsDummy.objects.create(title='ABC')
        youpi = ANDPermissionsDummy.objects.create(title='youpi woopaa')
        wonder = ANDPermissionsDummy.objects.create(title='A Wonderful World!!')
        plop = ANDPermissionsDummy.objects.create(title='plop')
        self.check_can_view('/andpermissionsdummys/', [wonder.urlid], 403)
        self.check_can_add('/andpermissionsdummys/', 403, field='title')
        self.check_can_change(wonder.urlid, 403, field='title')

        self.user.username = 'toto'
        self.user.save()
        self.check_can_view('/andpermissionsdummys/', [wonder.urlid])
        self.check_can_view_one(abc.urlid, 404)
        self.check_can_view_one(youpi.urlid, 404)
        self.check_can_view_one(plop.urlid, 404)
        self.check_can_add('/andpermissionsdummys/', 403, field='title')
        self.check_can_change(wonder.urlid, 403, field='title')
        self.check_can_change(youpi.urlid, 403, field='title')

    def test_or_permissions(self):
        self.authenticate()
        abc = ORPermissionsDummy.objects.create(title='ABC')
        youpi = ORPermissionsDummy.objects.create(title='youpi woopaa')
        wonder = ORPermissionsDummy.objects.create(title='A Wonderful World!!')
        plop = ORPermissionsDummy.objects.create(title='plop')
        self.check_can_view('/orpermissionsdummys/', [abc.urlid, wonder.urlid])
        self.check_can_add('/andpermissionsdummys/', 403, field='title')
        self.check_can_change(wonder.urlid, 403, field='title')

        self.user.username = 'toto'
        self.user.save()
        self.check_can_view('/orpermissionsdummys/', [abc.urlid, youpi.urlid, wonder.urlid])
        self.check_can_view_one(plop.urlid, 404)
        self.check_can_add('/orpermissionsdummys/', field='title')
        self.check_can_change(wonder.urlid, field='title')
        self.check_can_change(plop.urlid, 404, field='title')


import sys
import yaml

import django
from django.conf import settings as django_settings
from djangoldp.conf.ldpsettings import LDPSettings
from djangoldp.tests.server_settings import yaml_config

# load test config
config = yaml.safe_load(yaml_config)
ldpsettings = LDPSettings(config)
django_settings.configure(ldpsettings,
                          REST_FRAMEWORK = {
                                'DEFAULT_PAGINATION_CLASS': None
                          },
                          ANONYMOUS_USER_NAME=None)

django.setup()
from django.test.runner import DiscoverRunner

test_runner = DiscoverRunner(verbosity=1)

failures = test_runner.run_tests([
    # 'djangoldp.tests.tests_performance',
    'djangoldp.tests.tests_perf_get'
])
if failures:
    sys.exit(failures)



import logging

from rest_framework.views import APIView

from djangoldp.utils import is_authenticated_user
from djangoldp.views.commons import NoCSRFAuthentication

logger = logging.getLogger('djangoldp')


class LDPAPIView(APIView):
    '''extends rest framework APIView to support Solid standards'''
    authentication_classes = (NoCSRFAuthentication,)

    def dispatch(self, request, *args, **kwargs):
        '''overriden dispatch method to append some custom headers'''
        response = super().dispatch(request, *args, **kwargs)
        
        if response.status_code in [201, 200] and isinstance(response.data, dict) and '@id' in response.data:
            response["Location"] = str(response.data['@id'])
        else:
            pass

        if is_authenticated_user(request.user):
            try:
                response['User'] = request.user.urlid
            except AttributeError:
                pass


# Generated by Django 4.2.3 on 2023-08-31 15:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djangoldp', '0015_auto_20210125_1847'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='activity',
            options={'default_permissions': {'delete', 'change', 'view', 'add', 'control'}},
        ),
        migrations.AlterModelOptions(
            name='follower',
            options={'default_permissions': {'delete', 'change', 'view', 'add', 'control'}},
        ),
        migrations.AlterModelOptions(
            name='ldpsource',
            options={'default_permissions': {'delete', 'change', 'view', 'add', 'control'}, 'ordering': ('federation',)},
        ),
        migrations.AlterModelOptions(
            name='scheduledactivity',
            options={'default_permissions': {'delete', 'change', 'view', 'add', 'control'}},
        ),
    ]



from copy import copy
from django.conf import settings
from django.http import Http404
from rest_framework.permissions import BasePermission, DjangoObjectPermissions, OR, AND
from rest_framework.filters import BaseFilterBackend
from rest_framework_guardian.filters import ObjectPermissionsFilter
from djangoldp.filters import OwnerFilterBackend, NoFilterBackend, PublicFilterBackend, IPFilterBackend, ActiveFilterBackend
from djangoldp.utils import is_anonymous_user, is_authenticated_user, check_client_ip


DEFAULT_DJANGOLDP_PERMISSIONS = {'view', 'add', 'change', 'delete', 'control'}
DEFAULT_RESOURCE_PERMISSIONS = {'view', 'change', 'delete', 'control'}
DEFAULT_CONTAINER_PERMISSIONS = {'view', 'add'}

def join_filter_backends(*permissions_or_filters:BaseFilterBackend|BasePermission, model:object, union:bool=False) -> BaseFilterBackend:
    '''Creates a new Filter backend by joining a list of existing backends.
    It chains the filterings or joins them, depending on the argument union'''
    backends = []
    for permission_or_filter in permissions_or_filters:
        if hasattr(permission_or_filter, 'get_filter_backend'):
            backends.append(permission_or_filter.get_filter_backend(model))
        elif isinstance(permission_or_filter, type) and issubclass(permission_or_filter, BaseFilterBackend):
            backends.append(permission_or_filter)
    class JointFilterBackend(BaseFilterBackend):
        def __init__(self) -> None:
            self.filters = []
            for backend in backends:
                if backend:
                    self.filters.append(backend())
        def filter_queryset(self, request:object, queryset:object, view:object) -> object:
            if union:
                result = queryset.none() #starts with empty for union
            else:
                result = queryset
            for filter in self.filters:
                if union:
                    result = result | filter.filter_queryset(request, queryset, view)
                else:
                    result = filter.filter_queryset(request, result, view)
            return result
    return JointFilterBackend

permission_map ={
    'GET': ['%(app_label)s.view_%(model_name)s'],
    'OPTIONS': [],
    'HEAD': ['%(app_label)s.view_%(model_name)s'],
    'POST': ['%(app_label)s.add_%(model_name)s'],
    'PUT': ['%(app_label)s.change_%(model_name)s'],
    'PATCH': ['%(app_label)s.change_%(model_name)s'],
    'DELETE': ['%(app_label)s.delete_%(model_name)s'],
}

# Patch of OR and AND classes to enable chaining of LDPBasePermission
def OR_get_permissions(self, user, model, obj=None):
    perms1 = self.op1.get_permissions(user, model, obj) if hasattr(self.op1, 'get_permissions') else set()
    perms2 = self.op2.get_permissions(user, model, obj) if hasattr(self.op2, 'get_permissions') else set()
    return set.union(perms1, perms2)    
OR.get_permissions = OR_get_permissions
def OR_get_filter_backend(self, model):
    return join_filter_backends(self.op1, self.op2, model=model, union=True)
OR.get_filter_backend = OR_get_filter_backend
OR.__repr__ = lambda self: f"{self.op1}|{self.op2}"

def AND_get_permissions(self, user, model, obj=None):
    perms1 = self.op1.get_permissions(user, model, obj) if hasattr(self.op1, 'get_permissions') else set()
    perms2 = self.op2.get_permissions(user, model, obj) if hasattr(self.op2, 'get_permissions') else set()
    return set.intersection(perms1, perms2)    
AND.get_permissions = AND_get_permissions
def AND_get_filter_backend(self, model):
    return join_filter_backends(self.op1, self.op2, model=model, union=False)
AND.get_filter_backend = AND_get_filter_backend
AND.__repr__ = lambda self: f"{self.op1}&{self.op2}"

class LDPBasePermission(BasePermission):
    """
    A base class from which all permission classes should inherit.
    Extends the DRF permissions class to include the concept of model-permissions, separate from the view, and to
    change to a system of outputting permissions sets for the serialization of WebACLs
    """
    # filter backends associated with the permissions class. This will be used to filter queryset in the (auto-generated)
    # view for a model, and in the serializing nested fields
    filter_backend = ActiveFilterBackend
    # by default, all permissions
    permissions = getattr(settings, 'DJANGOLDP_PERMISSIONS', DEFAULT_DJANGOLDP_PERMISSIONS)
    # perms_map defines the permissions required for different methods
    perms_map = permission_map

    @classmethod
    def get_filter_backend(cls, model):
        '''returns the Filter backend associated with this permission class'''
        return cls.filter_backend
    def check_all_permissions(self, required_permissions):
        '''returns True if the all the permissions are included in the permissions of the class'''
        return all([permission.split('.')[1].split('_')[0] in self.permissions for permission in required_permissions])
    def get_allowed_methods(self):
        '''returns the list of methods allowed for the permissions of the class, depending on the permission map'''
        return [method for method, permissions in self.perms_map.items() if self.check_all_permissions(permissions)]
    def has_permission(self, request, view):
        '''checks if the request is allowed at all, based on its method and the permissions of the class'''
        return request.method in self.get_allowed_methods()
    def has_object_permission(self, request, view, obj=None):
        '''checks if the access to the object is allowed,'''
        return self.has_permission(request, view)
    def get_permissions(self, user, model, obj=None):
        '''returns the permissions the user has on a given model or on a given object'''
        return self.permissions.intersection(DEFAULT_RESOURCE_PERMISSIONS if obj else DEFAULT_CONTAINER_PERMISSIONS)

class AnonymousReadOnly(LDPBasePermission):
    """Anonymous users can only view, no check for others"""
    permissions = {'view'}
    def has_permission(self, request, view):
        return super().has_permission(request, view) or is_authenticated_user(request.user)
    def get_permissions(self, user, model, obj=None):
        if is_anonymous_user(user):
            return self.permissions
        else:
            return super().permissions #all permissions

class AuthenticatedOnly(LDPBasePermission):
    """Only authenticated users have permissions"""
    def has_permission(self, request, view):
        return request.method=='OPTIONS' or is_authenticated_user(request.user)

class ReadOnly(LDPBasePermission):
    """Users can only view"""
    permissions = {'view'}

class ReadAndCreate(LDPBasePermission):
    """Users can only view and create"""
    permissions = {'view', 'add'}

class CreateOnly(LDPBasePermission):
    """Users can only view and create"""
    permissions = {'add'}

class ACLPermissions(DjangoObjectPermissions, LDPBasePermission):
    """Permissions based on the rights given in db, on model for container requests or on object for resource requests"""
    filter_backend = ObjectPermissionsFilter
    perms_map = permission_map
    def has_permission(self, request, view):
        if view.action in ('list', 'create'): # The container permission only apply to containers requests
            return super().has_permission(request, view)
        return True

    def get_permissions(self, user, model, obj=None):
        model_name = model._meta.model_name
        app_label = model._meta.app_label
        if obj:
            return {perm.replace('_'+model_name, '') for perm in user.get_all_permissions(obj)}

        permissions = set(filter(lambda perm: perm.startswith(app_label) and perm.endswith(model_name), user.get_all_permissions()))
        return {perm.replace(app_label+'.', '').replace('_'+model_name, '') for perm in permissions}

class OwnerPermissions(LDPBasePermission):
    """Gives all permissions to the owner of the object"""
    filter_backend = OwnerFilterBackend
    def check_permission(self, user, model, obj):
        if user.is_superuser:
            return True
        if getattr(model._meta, 'owner_field', None):
            field = model._meta.get_field(model._meta.owner_field)
            if field.many_to_many or field.one_to_many:
                return user in getattr(obj, field.get_accessor_name()).all()
            else:
                return user == getattr(obj, model._meta.owner_field)
        if getattr(model._meta, 'owner_urlid_field', None) is not None:
            return is_authenticated_user(user) and user.urlid == getattr(obj, model._meta.owner_urlid_field)
        return True

    def has_object_permission(self, request, view, obj=None):
        return self.check_permission(request.user, view.model, obj)
    def get_permissions(self, user, model, obj=None):
        if not obj or self.check_permission(user, model, obj):
            return self.permissions
        return set()

class OwnerCreatePermission(LDPBasePermission):
    '''only accepts the creation of new resources if the owner of the created resource is the user of the request'''
    def check_patch(self, first, second, user):
        diff = first - second
        return diff == set() or diff == {user.urlid}

    def has_permission(self, request:object, view:object) -> bool:
        if request.method != 'POST':
            return super().has_permission(request, view)
        if is_anonymous_user(request.user):
            return False
        owner = None
        if getattr(view.model._meta, 'owner_field', None):
            field = view.model._meta.get_field(view.model._meta.owner_field)
            if field.many_to_many or field.one_to_many:
                owner = request.data[field.get_accessor_name()]
            else:
                owner = request.data[view.model._meta.owner_field]
        if getattr(view.model._meta, 'owner_urlid_field', None):
            owner = request.data[view.model._meta.owner_urlid_field]
        return not owner or owner['@id'] == request.user.urlid

class PublicPermission(LDPBasePermission):
    """Gives read-only access to resources which have a public flag to True"""
    filter_backend = PublicFilterBackend
    permissions = {'view', 'add'}
    def has_object_permission(self, request, view, obj=None):
        assert hasattr(view.model._meta, 'public_field'), \
            f'Model {view.model} has PublicPermission applied without "public_field" defined'
        public_field = view.model._meta.public_field

        if getattr(obj, public_field, False):
            return super().has_object_permission(request, view, obj)
        return False

class JoinMembersPermission(LDPBasePermission):
    filter_backend = None
    def has_permission(self, request:object, view:object) -> bool:
        if is_anonymous_user(request.user):
            return False
        return request.method == 'PATCH'

    def check_patch(self, first, second, user):
        diff = first - second
        return diff == set() or diff == {user.urlid}

    def has_object_permission(self, request:object, view:object, obj:object) -> bool:
        '''only accept patch request, only if the only difference on the user_set is the user'''
        if not self.has_permission(request, view) or not obj or not 'user_set' in request.data:
            return False
        new_members = request.data['user_set']
        if not isinstance(new_members, list):
            new_members = [new_members]
        new_ids = {user['@id'] for user in new_members}
        old_ids = {user.urlid for user in obj.members.user_set.all()}
        return self.check_patch(new_ids, old_ids, request.user) and self.check_patch(old_ids, new_ids, request.user)

    def get_permissions(self, user, model, obj=None):
        return set()


class IPOpenPermissions(LDPBasePermission):
    filter_backend = IPFilterBackend
    def has_permission(self, request, view):
        return check_client_ip(request)

    def has_object_permission(self, request, view, obj):
        return check_client_ip(request)

    def get_permissions(self, user, model, obj=None):
        #Will always say there is no migrations, not taking the IP into accounts
        return set()


class InheritPermissions(LDPBasePermission):
    """Gets the permissions from a related objects"""
    @classmethod
    def get_parent_fields(cls, model: object) -> list:
        '''checks that the model is adequately configured and returns the associated model'''
        assert hasattr(model._meta, 'inherit_permissions') and isinstance(model._meta.inherit_permissions, list), \
            f'Model {model} has InheritPermissions applied without "inherit_permissions" defined as a list'

        return model._meta.inherit_permissions

    @classmethod
    def get_parent_model(cls, model:object, field_name:str) -> object:
        parent_model = model._meta.get_field(field_name).related_model
        assert hasattr(parent_model._meta, 'permission_classes'), \
            f'Related model {parent_model} has no "permission_classes" defined'
        return parent_model

    def get_parent_objects(self, obj:object, field_name:str) -> list:
        '''gets the parent object'''
        if obj is None:
            return []
        field = obj._meta.get_field(field_name)
        if field.many_to_many or field.one_to_many:
            return getattr(obj, field.get_accessor_name()).all()
        parent = getattr(obj, field_name, None)
        return [parent] if parent else []
    
    @classmethod
    def clone_with_model(self, request:object, view:object, model:object) -> tuple:
        '''changes the model on the argument, so that they can be called on the parent model'''
        # For some reason if we copy the request itself, we go into an infinite loop, so take the native request instead
        _request = copy(request._request)
        _request.model = model
        _request.data = request.data #because the data is not present on the native request
        _request._request = _request #so that it can be nested
        _view = copy(view)
        _view.queryset = None #to make sure the model is taken into account
        _view.model = model
        return _request, _view

    @classmethod
    def generate_filter_backend(cls, parent:object, field_name:str) -> BaseFilterBackend:
        '''returns a new Filter backend that applies all filters of the parent model'''
        filter_arg = f'{field_name}__in'
        backends = {perm().get_filter_backend(parent) for perm in parent._meta.permission_classes}

        class InheritFilterBackend(BaseFilterBackend):
            def __init__(self) -> None:
                self.filters = []
                for backend in backends:
                    if backend:
                        self.filters.append(backend())
            def filter_queryset(self, request:object, queryset:object, view:object) -> object:
                request, view = InheritPermissions.clone_with_model(request, view, parent)
                for filter in self.filters:
                    allowed_parents = filter.filter_queryset(request, parent.objects.all(), view)
                    queryset = queryset.filter(**{filter_arg: allowed_parents})
                return queryset
        return InheritFilterBackend
    
    @classmethod
    def generate_filter_backend_for_none(cls, fields) -> BaseFilterBackend:
        '''returns a new Filter backend that checks that none of the parent fields are set'''
        class InheritNoneFilterBackend(BaseFilterBackend):
            def filter_queryset(self, request:object, queryset:object, view:object) -> object:
                return queryset.filter(**{field: None for field in fields})
        return InheritNoneFilterBackend

    @classmethod
    def get_filter_backend(cls, model:object) -> BaseFilterBackend:
        '''Returns a union filter backend of all filter backends of parents'''
        fields = cls.get_parent_fields(model)
        backends = [cls.generate_filter_backend(cls.get_parent_model(model, field), field) for field in fields]
        backend_none = cls.generate_filter_backend_for_none(fields)
        return join_filter_backends(*backends, backend_none, model=model, union=True)

    def has_permission(self, request:object, view:object) -> bool:
        '''Returns True unless we're trying to create a resource with a link to a parent we're not allowed to change'''
        if request.method == 'POST':
            for field in InheritPermissions.get_parent_fields(view.model):
                if field in request.data:
                    model = InheritPermissions.get_parent_model(view.model, field)
                    parent = model.objects.get(urlid=request.data[field]['@id'])
                    _request, _view = InheritPermissions.clone_with_model(request, view, model)
                    if not all([perm().has_object_permission(_request, _view, parent) for perm in model._meta.permission_classes]):
                        return False
        return True
    
    def has_object_permission(self, request:object, view:object, obj:object) -> bool:
        '''Returns True if at least one inheriting object has permission'''
        if not obj:
            return super().has_object_permission(request, view, obj)
        parents = []
        for field in InheritPermissions.get_parent_fields(view.model):
            model = InheritPermissions.get_parent_model(view.model, field)
            parent_request, parent_view = InheritPermissions.clone_with_model(request, view, model)
            for parent_object in self.get_parent_objects(obj, field):
                parents.append(parent_object)
                try:
                    if all([perm().has_object_permission(parent_request, parent_view, parent_object) for perm in model._meta.permission_classes]):
                        return True
                except Http404:
                    #keep trying
                    pass
        # return False if there were parent resources but none accepted
        return False if parents else True
    
    def get_permissions(self, user:object, model:object, obj:object=None) -> set:
        '''returns a union of all inheriting linked permissions'''
        perms = set()
        parents = []
        for field in InheritPermissions.get_parent_fields(model):
            parent_model = InheritPermissions.get_parent_model(model, field)
            for parent_object in self.get_parent_objects(obj, field):
                parents.append(parent_object)
                perms = perms.union(set.intersection(*[perm().get_permissions(user, parent_model, parent_object) 
                                               for perm in parent_model._meta.permission_classes]))
        if parents:
            return perms
        return super().get_permissions(user, model, obj)


from django.conf import settings
from guardian.utils import get_anonymous_user

PASSTHROUGH_IPS = getattr(settings, 'PASSTHROUGH_IPS', '')

# convenience function returns True if user is anonymous
def is_anonymous_user(user):
    anonymous_username = getattr(settings, 'ANONYMOUS_USER_NAME', None)
    return user.is_anonymous or user.username == anonymous_username


# convenience function returns True if user is authenticated
def is_authenticated_user(user):
    anonymous_username = getattr(settings, 'ANONYMOUS_USER_NAME', None)
    return user.is_authenticated and user.username != anonymous_username


# this method is used to check if a given IP is part of the PASSTHROUGH_IPS list
def check_client_ip(request):
    x_forwarded_for = request.headers.get('x-forwarded-for')
    if x_forwarded_for:
        if any(ip in x_forwarded_for.replace(' ', '').split(',') for ip in PASSTHROUGH_IPS):
            return True
    elif request.META.get('REMOTE_ADDR') in PASSTHROUGH_IPS:
        return True
    return False



# Generated by Django 4.2.3 on 2023-09-03 20:26

from django.db import migrations
import djangoldp.fields


class Migration(migrations.Migration):

    dependencies = [
        ('djangoldp', '0016_alter_activity_options_alter_follower_options_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='activity',
            name='urlid',
            field=djangoldp.fields.LDPUrlField(blank=True, db_index=True, null=True, unique=True),
        ),
        migrations.AlterField(
            model_name='follower',
            name='urlid',
            field=djangoldp.fields.LDPUrlField(blank=True, db_index=True, null=True, unique=True),
        ),
        migrations.AlterField(
            model_name='ldpsource',
            name='urlid',
            field=djangoldp.fields.LDPUrlField(blank=True, db_index=True, null=True, unique=True),
        ),
    ]



import json
import logging

from django.core.exceptions import ObjectDoesNotExist
from django.db import IntegrityError, transaction
from django.http import Http404
from django.views import View

from rest_framework import status
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.utils import model_meta
from rest_framework.views import APIView

import validators

from djangoldp.activities import (
    ACTIVITY_SAVING_SETTING,
    ActivityPubService,
    ActivityQueueService,
    as_activitystream,
)
from djangoldp.activities.errors import (
    ActivityStreamDecodeError,
    ActivityStreamValidationError,
)
from djangoldp.models import Follower, Model

logger = logging.getLogger('djangoldp')


class InboxView(APIView):
    """
    Receive linked data notifications
    """
    permission_classes = [AllowAny, ]

    def post(self, request, *args, **kwargs):
        '''
        receiver for inbox messages. See https://www.w3.org/TR/ldn/
        '''
        try:
            activity = json.loads(request.body, object_hook=as_activitystream)
            activity.validate()
        except ActivityStreamDecodeError:
            return Response('Activity type unsupported', status=status.HTTP_405_METHOD_NOT_ALLOWED)
        except ActivityStreamValidationError as e:
            return Response(str(e), status=status.HTTP_400_BAD_REQUEST)

        try:
            self._handle_activity(activity, **kwargs)
        except IntegrityError:
            return Response({'Unable to save due to an IntegrityError in the receiver model'},
                            status=status.HTTP_200_OK)
        except ValueError as e:
            return Response(str(e), status=status.HTTP_400_BAD_REQUEST)

        # save the activity and return 201
        if ACTIVITY_SAVING_SETTING == 'VERBOSE':
            obj = ActivityQueueService._save_sent_activity(activity.to_json(), local_id=request.path_info, success=True,
                                                           type=activity.type)

            response = Response({}, status=status.HTTP_201_CREATED)
            response['Location'] = obj.urlid
        
        else:
            response = Response({}, status=status.HTTP_200_OK)

        return response

    def _handle_activity(self, activity, **kwargs):
        if activity.type == 'Add':
            self.handle_add_activity(activity, **kwargs)
        elif activity.type == 'Remove':
            self.handle_remove_activity(activity, **kwargs)
        elif activity.type == 'Delete':
            self.handle_delete_activity(activity, **kwargs)
        elif activity.type == 'Create' or activity.type == 'Update':
            self.handle_create_or_update_activity(activity, **kwargs)
        elif activity.type == 'Follow':
            self.handle_follow_activity(activity, **kwargs)

    def atomic_get_or_create_nested_backlinks(self, obj, object_model=None, update=False):
        '''
        a version of get_or_create_nested_backlinks in which all nested backlinks are created, or none of them are
        '''
        try:
            with transaction.atomic():
                return self._get_or_create_nested_backlinks(obj, object_model, update)
        except IntegrityError as e:
            logger.error(str(e))
            logger.warning(
                'received a backlink which you were not able to save because of a constraint on the model field.')
            raise e

    def _get_or_create_nested_backlinks(self, obj, object_model=None, update=False):
        '''
        recursively deconstructs a tree of nested objects, using get_or_create on each leaf/branch
        :param obj: Dict representation of the object
        :param object_model: The Model class of the object. Will be discovered if set to None
        :param update: if True will update retrieved objects with new data
        :raises Exception: if get_or_create fails on a branch, the creation will be reversed and the Exception re-thrown
        '''
        # store a list of the object's sub-items
        if object_model is None:
            object_model = Model.get_subclass_with_rdf_type(obj['@type'])
        if object_model is None:
            raise Http404('unable to store type ' + obj['@type'] + ', model with this rdf_type not found')
        branches = {}

        for item in obj.items():
            # TODO: parse other data types. Match the key to the field_name
            if isinstance(item[1], dict):
                item_value = item[1]
                item_model = Model.get_subclass_with_rdf_type(item_value['@type'])
                if item_model is None:
                    raise Http404(
                        'unable to store type ' + item_value['@type'] + ', model with this rdf_type not found')

                # push nested object tuple as a branch
                backlink = self._get_or_create_nested_backlinks(item_value, item_model)
                branches[item[0]] = backlink

        # get or create the backlink
        try:
            if obj['@id'] is None or not validators.url(obj['@id']):
                raise ValueError('received invalid urlid ' + str(obj['@id']))
            external = Model.get_or_create_external(object_model, obj['@id'], update=update, **branches)

            # creating followers, to inform distant resource of changes to local connection
            if Model.is_external(external):
                # this is handled with Followers, where each local child of the branch is followed by its external parent
                for item in obj.items():
                    urlid = item[1]
                    if isinstance(item[1], dict):
                        urlid = urlid['@id']
                    if not isinstance(urlid, str) or not validators.url(urlid):
                        continue

                    if not Model.is_external(urlid):
                        ActivityPubService.save_follower_for_target(external.urlid, urlid)

            return external

        # this will be raised when the object was local, but it didn't exist
        except ObjectDoesNotExist:
            raise Http404(getattr(object_model._meta, 'label', 'Unknown Model') + ' ' + str(obj['@id']) + ' does not exist')

    # TODO: a fallback here? Saving the backlink as Object or similar
    def _get_subclass_with_rdf_type_or_404(self, rdf_type):
        model = Model.get_subclass_with_rdf_type(rdf_type)
        if model is None:
            raise Http404('unable to store type ' + rdf_type + ', model not found')
        return model

    def handle_add_activity(self, activity, **kwargs):
        '''
        handles Add Activities. See https://www.w3.org/ns/activitystreams
        Indicates that the actor has added the object to the target
        '''
        object_model = self._get_subclass_with_rdf_type_or_404(activity.object['@type'])
        target_model = self._get_subclass_with_rdf_type_or_404(activity.target['@type'])

        try:
            target = target_model.objects.get(urlid=activity.target['@id'])
        except target_model.DoesNotExist:
            return Response({}, status=status.HTTP_404_NOT_FOUND)

        # store backlink(s) in database
        backlink = self.atomic_get_or_create_nested_backlinks(activity.object, object_model)

        # add object to target
        target_info = model_meta.get_field_info(target_model)

        for field_name, relation_info in target_info.relations.items():
            if relation_info.related_model == object_model:
                attr = getattr(target, field_name)
                if not attr.filter(urlid=backlink.urlid).exists():
                    attr.add(backlink)
                    ActivityPubService.save_follower_for_target(backlink.urlid, target.urlid)

    def handle_remove_activity(self, activity, **kwargs):
        '''
        handles Remove Activities. See https://www.w3.org/ns/activitystreams
        Indicates that the actor has removed the object from the origin
        '''
        # TODO: Remove Activity may pass target instead
        object_model = self._get_subclass_with_rdf_type_or_404(activity.object['@type'])
        origin_model = self._get_subclass_with_rdf_type_or_404(activity.origin['@type'])

        # get the model reference to saved object
        try:
            origin = origin_model.objects.get(urlid=activity.origin['@id'])
            object_instance = object_model.objects.get(urlid=activity.object['@id'])
        except origin_model.DoesNotExist:
            raise Http404(activity.origin['@id'] + ' did not exist')
        except object_model.DoesNotExist:
            return

        # remove object from origin
        origin_info = model_meta.get_field_info(origin_model)

        for field_name, relation_info in origin_info.relations.items():
            if relation_info.related_model == object_model:
                attr = getattr(origin, field_name)
                if attr.filter(urlid=object_instance.urlid).exists():
                    attr.remove(object_instance)
                    ActivityPubService.remove_followers_for_resource(origin.urlid, object_instance.urlid)

    def handle_create_or_update_activity(self, activity, **kwargs):
        '''
        handles Create & Update Activities. See https://www.w3.org/ns/activitystreams
        '''
        object_model = self._get_subclass_with_rdf_type_or_404(activity.object['@type'])
        self.atomic_get_or_create_nested_backlinks(activity.object, object_model, update=True)

    def handle_delete_activity(self, activity, **kwargs):
        '''
        handles Remove Activities. See https://www.w3.org/ns/activitystreams
        Indicates that the actor has deleted the object
        '''
        object_model = self._get_subclass_with_rdf_type_or_404(activity.object['@type'])

        # get the model reference to saved object
        try:
            object_instance = object_model.objects.get(urlid=activity.object['@id'])
        except object_model.DoesNotExist:
            return

        # disable backlinks first - prevents a duplicate being sent back
        object_instance.allow_create_backlink = False
        object_instance.save()
        object_instance.delete()
        urlid = getattr(object_instance, 'urlid', None)
        if urlid is not None:
            for follower in Follower.objects.filter(follower=urlid):
                follower.delete()

    def handle_follow_activity(self, activity, **kwargs):
        '''
        handles Follow Activities. See https://www.w3.org/ns/activitystreams
        Indicates that the actor is following the object, and should receive Updates on what happens to it
        '''
        object_model = self._get_subclass_with_rdf_type_or_404(activity.object['@type'])

        # get the model reference to saved object
        try:
            object_instance = object_model.objects.get(urlid=activity.object['@id'])
        except object_model.DoesNotExist:
            raise Http404(activity.object['@id'] + ' did not exist')
        if Model.is_external(object_instance):
            raise Http404(activity.object['@id'] + ' is not local to this server')

        # get the inbox field from the actor
        if isinstance(activity.actor, str):
            inbox = activity.actor
        else:
            inbox = getattr(activity.actor, 'inbox', None)
            if inbox is None:
                inbox = getattr(activity.actor, 'id', getattr(activity.actor, '@id'))

        if not Follower.objects.filter(object=object_instance.urlid, inbox=inbox).exists():
            Follower.objects.create(object=object_instance.urlid, inbox=inbox)
